# Unit test backtesting_stats
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import Ridge
from skforecast.stats import Sarimax
from skforecast.recursive import ForecasterRecursive
from skforecast.recursive import ForecasterStats
from skforecast.model_selection._split import TimeSeriesFold
from skforecast.model_selection import backtesting_stats

# Fixtures
from ....recursive.tests.tests_forecaster_stats.fixtures_forecaster_stats import y_datetime
from ....recursive.tests.tests_forecaster_stats.fixtures_forecaster_stats import exog_datetime


def test_backtesting_forecaster_TypeError_when_forecaster_not_supported_types():
    """
    Test TypeError is raised in backtesting_forecaster if Forecaster is not one 
    of the types supported by the function.
    """
    forecaster = ForecasterRecursive(
                    estimator = Ridge(random_state=123),
                    lags      = 2
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )

    err_msg = re.escape(
        "`forecaster` must be of type `ForecasterStats`, for all other "
        "types of forecasters use the functions available in the other "
        "`model_selection` modules."
    )
    with pytest.raises(TypeError, match = err_msg):
        backtesting_stats(
            forecaster    = forecaster,
            y             = y_datetime,
            cv            = cv,
            metric        = 'mean_absolute_error',
            exog          = None,
            alpha         = None,
            interval      = None,
            verbose       = False,
            show_progress = False
        )


@pytest.mark.parametrize("n_jobs", [1, -1, 'auto'],
                         ids=lambda n: f'n_jobs: {n}')
def test_output_backtesting_stats_no_refit_no_exog_no_remainder_with_mocked(n_jobs):
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, no exog, 
    no refit, 12 observations to backtest, steps=3 (no remainder), metric='mean_squared_error'. 
    Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                       forecaster = forecaster,
                                       y          = y_datetime,
                                       cv         = cv,
                                       metric     = 'mean_squared_error',
                                       alpha      = None,
                                       interval   = None,
                                       n_jobs     = n_jobs,
                                       verbose    = True
                                   )
    
    expected_metric = pd.DataFrame({"mean_squared_error": [0.03683793335495359]})
    expected_preds = pd.DataFrame(
        data    = np.array([0.51853756, 0.5165776 , 0.51790214, 0.80140303, 0.84979734,
                            0.91918321, 0.84363512, 0.8804787 , 0.91651026, 0.42747836,
                            0.39041178, 0.23407875]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


@pytest.mark.parametrize("initial_train_size", 
                         [len(y_datetime) - 12, "2037-12-31"],
                         ids=lambda init: f'initial_train_size: {init}')
def test_output_backtesting_stats_no_refit_no_exog_remainder_with_mocked(initial_train_size):
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, no exog, 
    yes refit, 12 observations to backtest, steps=5 (remainder), metric='mean_squared_error'. 
    Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 5,
             initial_train_size    = initial_train_size,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                       forecaster = forecaster,
                                       y          = y_datetime,
                                       cv         = cv,
                                       metric     = 'mean_squared_error',
                                       alpha      = None,
                                       interval   = None,
                                       verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_squared_error": [0.07396344749165738]})
    expected_preds = pd.DataFrame(
        data    = np.array([0.51853756, 0.5165776 , 0.51790214, 0.51193703, 0.4991191 ,
                            0.89343704, 0.95023804, 1.00278782, 1.07322123, 1.13932909,
                            0.5673885 , 0.43713008]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


@pytest.mark.parametrize("n_jobs", [1, -1, 'auto'],
                         ids=lambda n: f'n_jobs: {n}')
def test_output_backtesting_stats_yes_refit_no_exog_no_remainder_with_mocked(n_jobs):
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, 
    no exog, yes refit, 12 observations to backtest, steps=3 (no remainder), 
    metric='mean_squared_error'. (Mocked done with skforecast 0.7.0.)
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = True,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                       forecaster = forecaster,
                                       y          = y_datetime,
                                       cv         = cv,
                                       metric     = 'mean_squared_error',
                                       alpha      = None,
                                       interval   = None,
                                       n_jobs     = n_jobs,
                                       verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_squared_error": [0.038704200731126036]})
    expected_preds = pd.DataFrame(
        data    = np.array([0.51853756, 0.5165776 , 0.51790214, 0.80295192, 0.85238217,
                            0.9244119 , 0.84173367, 0.8793909 , 0.91329115, 0.42336972,
                            0.38434305, 0.2093133 ]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


@pytest.mark.parametrize("n_jobs", [1, -1, "auto"],
                         ids=lambda n: f'n_jobs: {n}')
def test_output_backtesting_stats_yes_refit_no_exog_remainder_with_mocked(n_jobs):
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, 
    no exog, yes refit, 12 observations to backtest, steps=5 (remainder), 
    metric='mean_squared_error'. (Mocked done with skforecast 0.7.0.)
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 5,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = True,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        alpha      = None,
                                        interval   = None,
                                        n_jobs     = n_jobs,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_squared_error": [0.0754085450012623]})
    expected_preds = pd.DataFrame(
        data    = np.array([0.51853756, 0.5165776 , 0.51790214, 0.51193703, 0.4991191 ,
                            0.89513678, 0.94913026, 1.00437767, 1.07534674, 1.14049886,
                            0.56289528, 0.40343592]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


def test_output_backtesting_stats_yes_refit_fixed_train_size_no_exog_no_remainder_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, no exog, 
    yes refit, fixed_train_size yes, 12 observations to backtest, steps=3 (no remainder), 
    metric='mean_squared_error'. Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = True,
             fixed_train_size      = True,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_squared_error": [0.04116499283290456]})
    expected_preds = pd.DataFrame(
        data    = np.array([0.51853756, 0.5165776 , 0.51790214, 0.80320348, 0.85236718,
                            0.92421562, 0.85060945, 0.88539784, 0.92172861, 0.41776604,
                            0.37024487, 0.1878739 ]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


def test_output_backtesting_stats_yes_refit_fixed_train_size_no_exog_remainder_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, no exog, 
    yes refit, fixed_train_size yes, 12 observations to backtest, steps=5 (remainder), 
    metric='mean_squared_error'. Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 5,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = True,
             fixed_train_size      = True,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_squared_error": [0.07571810495568278]})
    expected_preds = pd.DataFrame(
        data    = np.array([0.51853756, 0.5165776 , 0.51790214, 0.51193703, 0.4991191 ,
                            0.8959923 , 0.95147449, 1.00612185, 1.07723486, 1.14356597,
                            0.56321268, 0.40920121]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)
    

def test_output_backtesting_stats_no_refit_yes_exog_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    no refit, 12 observations to backtest, steps=3 (no remainder), metric='mean_squared_error'. 
    Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        exog       = exog_datetime,
                                        metric     = 'mean_squared_error',
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_squared_error": [0.18551856781581755]})
    expected_preds = pd.DataFrame(
        data    = np.array([ 0.59409098,  0.78323365,  0.99609033,  0.87882152,  1.02722143,
                             1.16176993,  0.85860472,  0.86636317,  0.68987477,  0.17788782,
                            -0.13577   , -0.50570715]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


def test_output_backtesting_stats_yes_refit_yes_exog_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    yes refit, 12 observations to backtest, steps=3 (no remainder), metric='mean_squared_error'. 
    Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = True,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        exog       = exog_datetime,
                                        metric     = 'mean_squared_error',
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_squared_error": [0.198652574804823]})
    expected_preds = pd.DataFrame(
        data    = np.array([ 0.59409098,  0.78323365,  0.99609033,  0.8786089 ,  1.02218448,
                             1.15283534,  0.8597644 ,  0.87093769,  0.71221024,  0.16839089,
                            -0.16421948, -0.55386343]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


def test_output_backtesting_stats_yes_refit_fixed_train_size_yes_exog_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    yes refit, fixed_train_size, 12 observations to backtest, steps=5 (remainder), 
    metric='mean_squared_error'. Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 5,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = True,
             fixed_train_size      = True,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        exog       = exog_datetime,
                                        metric     = 'mean_squared_error',
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_squared_error": [0.0917642049564646]})
    expected_preds = pd.DataFrame(
        data    = np.array([0.59409098, 0.78323365, 0.99609033, 1.21449931, 1.4574755 ,
                            0.89448353, 0.99712901, 1.05090061, 0.92362208, 0.76795064,
                            0.45382855, 0.26823527]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


def my_metric(y_true, y_pred):  # pragma: no cover
    """
    Callable metric
    """
    metric = ((y_true - y_pred) / len(y_true)).mean()
    
    return metric


def test_output_backtesting_stats_no_refit_yes_exog_callable_metric_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    no refit, 12 observations to backtest, steps=3 (no remainder), callable metric. 
    Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        exog       = exog_datetime,
                                        metric     = my_metric,
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"my_metric": [0.007364452865679387]})
    expected_preds = pd.DataFrame(
        data    = np.array([ 0.59409098,  0.78323365,  0.99609033,  0.87882152,  1.02722143,
                             1.16176993,  0.85860472,  0.86636317,  0.68987477,  0.17788782,
                            -0.13577   , -0.50570715]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


def test_output_backtesting_stats_no_refit_no_exog_list_of_metrics_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    no refit, 12 observations to backtest, steps=3 (no remainder), list of metrics. 
    Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        metric     = [my_metric, 'mean_absolute_error'],
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"my_metric": [0.004423392707787538], 
                                    "mean_absolute_error": [0.1535720350789038]})
    expected_preds = pd.DataFrame(
        data    = np.array([0.51853756, 0.5165776 , 0.51790214, 0.80140303, 0.84979734,
                            0.91918321, 0.84363512, 0.8804787 , 0.91651026, 0.42747836,
                            0.39041178, 0.23407875]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


def test_output_backtesting_stats_yes_refit_no_exog_callable_metric_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, no exog, 
    yes refit, 12 observations to backtest, steps=3 (no remainder), callable metric. 
    Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = True,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        metric     = my_metric,
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"my_metric": [0.004644148042633733]})
    expected_preds = pd.DataFrame(
        data    = np.array([0.51853756, 0.5165776 , 0.51790214, 0.80295192, 0.85238217,
                            0.9244119 , 0.84173367, 0.8793909 , 0.91329115, 0.42336972,
                            0.38434305, 0.2093133 ]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


def test_output_backtesting_stats_yes_refit_fixed_train_size_yes_exog_list_of_metrics_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    yes refit, fixed_train_size, 12 observations to backtest, steps=3 (no remainder), 
    list of metrics. Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = True,
             fixed_train_size      = True,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        exog       = exog_datetime,
                                        metric     = [my_metric, 'mean_absolute_scaled_error'],
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"my_metric": [0.007877420102652216], 
                                    "mean_absolute_scaled_error": [3.899618814139531]})
    expected_preds = pd.DataFrame(
        data    = np.array([ 0.59409098,  0.78323365,  0.99609033,  0.88202026,  1.03241114,
                             1.16808941,  0.86535534,  0.87277596,  0.69357041,  0.16628876,
                            -0.17189178, -0.56342057]),
        columns = ['pred'],
        index   = pd.date_range(start='2038', periods=12, freq='YE')
    )
    expected_preds.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)
    

@pytest.mark.parametrize("alpha, interval", 
                         [(0.05, [1, 99]), 
                          (None, [2.5, 97.5])], 
                         ids = lambda values: f'alpha, interval: {values}')
def test_output_backtesting_stats_no_refit_yes_exog_interval_with_mocked(alpha, interval):
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    no refit, 12 observations to backtest, steps=3 (no remainder), metric='mean_absolute_error',
    interval. Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        exog       = exog_datetime,
                                        metric     = ['mean_absolute_error'],
                                        alpha      = alpha,
                                        interval   = interval,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_absolute_error": [0.3040748056175932]})
    expected_values = np.array([[ 0.59409098, -1.08941968,  2.27760163],
                                [ 0.78323365, -2.51570095,  4.08216824],
                                [ 0.99609033, -4.02240666,  6.01458732],
                                [ 0.87882152, -0.80468913,  2.56233218],
                                [ 1.02722143, -2.27171317,  4.32615603],
                                [ 1.16176993, -3.85672705,  6.18026692],
                                [ 0.85860472, -0.82490594,  2.54211537],
                                [ 0.86636317, -2.43257143,  4.16529777],
                                [ 0.68987477, -4.32862221,  5.70837176],
                                [ 0.17788782, -1.50562283,  1.86139848],
                                [-0.13577   , -3.4347046 ,  3.16316459],
                                [-0.50570715, -5.52420413,  4.51278984]])

    expected_backtest_predictions = pd.DataFrame(
                                        data    = expected_values,
                                        columns = ['pred', 'lower_bound', 'upper_bound'],
                                        index   = pd.date_range(start='2038', periods=12, freq='YE')
                                    )  
    expected_backtest_predictions.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])                                                   

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_backtest_predictions, backtest_predictions, atol=0.0001)


@pytest.mark.parametrize("alpha, interval", 
                         [(0.05, [1, 99]), 
                          (None, [2.5, 97.5])], 
                         ids = lambda values: f'alpha, interval: {values}')
def test_output_backtesting_stats_yes_refit_yes_exog_interval_with_mocked(alpha, interval):
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    yes refit, 12 observations to backtest, steps=3 (no remainder), 
    metric='mean_absolute_error', interval. Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = True,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        exog       = exog_datetime,
                                        metric     = 'mean_absolute_error',
                                        alpha      = alpha,
                                        interval   = interval,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_absolute_error": [0.31145145397674484]})
    expected_values = np.array([[ 0.59409098, -1.08941968,  2.27760163],
                                [ 0.78323365, -2.51570095,  4.08216824],
                                [ 0.99609033, -4.02240666,  6.01458732],
                                [ 0.8786089 , -0.8166216 ,  2.57383939],
                                [ 1.02218448, -2.25140783,  4.29577679],
                                [ 1.15283534, -3.7591572 ,  6.06482787],
                                [ 0.8597644 , -0.84109595,  2.56062475],
                                [ 0.87093769, -2.39317559,  4.13505098],
                                [ 0.71221024, -4.21947832,  5.64389879],
                                [ 0.16839089, -1.52752939,  1.86431116],
                                [-0.16421948, -3.50117442,  3.17273546],
                                [-0.55386343, -5.6847384 ,  4.57701154]])

    expected_backtest_predictions = pd.DataFrame(
                                        data    = expected_values,
                                        columns = ['pred', 'lower_bound', 'upper_bound'],
                                        index   = pd.date_range(start='2038', periods=12, freq='YE')
                                    )  
    expected_backtest_predictions.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])                                                     

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_backtest_predictions, backtest_predictions, atol=0.0001)


@pytest.mark.parametrize("alpha, interval", 
                         [(0.05, [1, 99]), 
                          (None, [2.5, 97.5])], 
                         ids = lambda values: f'alpha, interval: {values}')
def test_output_backtesting_stats_yes_refit_fixed_train_size_yes_exog_interval_with_mocked(alpha, interval):
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    yes refit, fixed_train_size, 12 observations to backtest, steps=3 (no remainder), 
    metric='mean_absolute_error', interval. Mocked done with skforecast 0.7.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = True,
             fixed_train_size      = True,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                       forecaster = forecaster,
                                       y          = y_datetime,
                                       cv         = cv,
                                       exog       = exog_datetime,
                                       metric     = ['mean_absolute_error'],
                                       alpha      = alpha,
                                       interval   = interval,
                                       verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_absolute_error": [0.31329767191681507]})
    expected_values = np.array([[ 0.59409098, -1.08941968,  2.27760163],
                                [ 0.78323365, -2.51570095,  4.08216824],
                                [ 0.99609033, -4.02240666,  6.01458732],
                                [ 0.88202026, -0.79712278,  2.56116331],
                                [ 1.03241114, -2.26295601,  4.32777829],
                                [ 1.16808941, -3.82752621,  6.16370504],
                                [ 0.86535534, -0.82088676,  2.55159744],
                                [ 0.87277596, -2.39786807,  4.14342   ],
                                [ 0.69357041, -4.32350585,  5.71064667],
                                [ 0.16628876, -1.53370994,  1.86628747],
                                [-0.17189178, -3.5242321 ,  3.18044855],
                                [-0.56342057, -5.74734656,  4.62050543]])

    expected_backtest_predictions = pd.DataFrame(
                                        data    = expected_values,
                                        columns = ['pred', 'lower_bound', 'upper_bound'],
                                        index   = pd.date_range(start='2038', periods=12, freq='YE')
                                    )
    expected_backtest_predictions.insert(0, 'fold', [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])                                                    

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_backtest_predictions, backtest_predictions, atol=0.0001)


def test_output_backtesting_stats_fold_stride_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    yes refit, fixed_train_size, 12 observations to backtest, steps=3 (no remainder), 
    fold stride. Mocked done with skforecast 0.18.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 6,
             initial_train_size    = len(y_datetime) - 12,
             fold_stride           = 3,
             refit                 = True,
             fixed_train_size      = True,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        exog       = exog_datetime,
                                        metric     = 'mean_absolute_error',
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_absolute_error": [0.313297672064354]})
    expected_preds = pd.DataFrame(
        data    = np.array([
                      [0.        ,  0.59409098],
                      [0.        ,  0.78323365],
                      [0.        ,  0.99609033],
                      [0.        ,  1.21449931],
                      [0.        ,  1.4574755 ],
                      [0.        ,  1.68696099],
                      [1.        ,  0.88202026],
                      [1.        ,  1.03241114],
                      [1.        ,  1.16808941],
                      [1.        ,  1.34390467],
                      [1.        ,  1.4696968 ],
                      [1.        ,  1.41286359],
                      [2.        ,  0.86535534],
                      [2.        ,  0.87277596],
                      [2.        ,  0.69357041],
                      [2.        ,  0.48557182],
                      [2.        ,  0.28762654],
                      [2.        ,  0.05433193],
                      [3.        ,  0.16628876],
                      [3.        , -0.17189178],
                      [3.        , -0.56342057]]),
        columns = ['fold', 'pred'],
        index   = pd.DatetimeIndex([
                    '2038-12-31', '2039-12-31', '2040-12-31', '2041-12-31',
                    '2042-12-31', '2043-12-31', '2041-12-31', '2042-12-31',
                    '2043-12-31', '2044-12-31', '2045-12-31', '2046-12-31',
                    '2044-12-31', '2045-12-31', '2046-12-31', '2047-12-31',
                    '2048-12-31', '2049-12-31', '2047-12-31', '2048-12-31',
                    '2049-12-31'
                ])
    ).astype({'fold': int})

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)


def test_output_backtesting_stats_fold_stride_greater_than_steps_with_mocked():
    """
    Test output of backtesting_stats with backtesting mocked, Series y is mocked, yes exog, 
    yes refit, fixed_train_size, 12 observations to backtest, steps=3 (no remainder), 
    fold stride greater than steps. Mocked done with skforecast 0.18.0.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 2,
             initial_train_size    = len(y_datetime) - 12,
             fold_stride           = 4,
             refit                 = True,
             fixed_train_size      = True,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric, backtest_predictions = backtesting_stats(
                                        forecaster = forecaster,
                                        y          = y_datetime,
                                        cv         = cv,
                                        exog       = exog_datetime,
                                        metric     = 'mean_absolute_error',
                                        alpha      = None,
                                        interval   = None,
                                        verbose    = False
                                   )
    
    expected_metric = pd.DataFrame({"mean_absolute_error": [0.080745243332592]})
    expected_preds = pd.DataFrame(
        data    = np.array([
                      [0.        , 0.59409098],
                      [0.        , 0.78323365],
                      [1.        , 0.76895405],
                      [1.        , 0.81593985],
                      [2.        , 0.77460313],
                      [2.        , 0.5839772 ]]),
        columns = ['fold', 'pred'],
        index   = pd.DatetimeIndex([
                    '2038-12-31', '2039-12-31', '2042-12-31', '2043-12-31',
                    '2046-12-31', '2047-12-31'
                ])
    ).astype({'fold': int})

    pd.testing.assert_frame_equal(expected_metric, metric, atol=0.0001)
    pd.testing.assert_frame_equal(expected_preds, backtest_predictions, atol=0.0001)



# Unit test _recursive_predict ForecasterRecursiveClassifier
# ==============================================================================
import numpy as np
import pandas as pd
from lightgbm import LGBMClassifier
from sklearn.linear_model import LogisticRegression
from skforecast.preprocessing import RollingFeaturesClassification
from skforecast.recursive import ForecasterRecursiveClassifier

# Fixtures
from .fixtures_forecaster_recursive_classifier import y
from .fixtures_forecaster_recursive_classifier import exog
from .fixtures_forecaster_recursive_classifier import exog_predict


def test_recursive_predict_output_when_estimator_is_LogisticRegression():
    """
    Test _recursive_predict output when using LogisticRegression as estimator.
    """
    y_dummy = pd.Series(
        np.array(['a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c']), 
        index=pd.date_range("2020-01-01", periods=15),
        name='y'
    )
    
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    forecaster.fit(y=y_dummy)

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=5)
    )
    predictions = forecaster._recursive_predict(
                      steps              = 5,
                      last_window_values = last_window_values,
                      exog_values        = exog_values
                  )
    
    expected = np.array([0., 1., 2., 0., 1.])
    
    np.testing.assert_array_almost_equal(predictions, expected)


def test_recursive_predict_output_when_estimator_is_LogisticRegression_and_exog():
    """
    Test _recursive_predict output when using LogisticRegression as estimator.
    """
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    forecaster.fit(y=y, exog=exog)

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=5, exog=exog_predict)
    )
    predictions = forecaster._recursive_predict(
                      steps              = 5,
                      last_window_values = last_window_values,
                      exog_values        = exog_values
                  )
    
    expected = np.array([1., 1., 1., 1., 1.])
    
    np.testing.assert_array_almost_equal(predictions, expected)


def test_recursive_predict_output_with_window_features():
    """
    Test _recursive_predict output with window features.
    """
    rolling = RollingFeaturesClassification(
        stats=['proportion', 'entropy'], window_sizes=4
    )
    forecaster = ForecasterRecursiveClassifier(
        LGBMClassifier(verbose=-1), lags=3, window_features=rolling
    )
    forecaster.fit(y=y, exog=exog)

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=10, exog=exog_predict)
    )
    predictions = forecaster._recursive_predict(
                      steps              = 10,
                      last_window_values = last_window_values,
                      exog_values        = exog_values
                  )
    
    expected = np.array([1., 1., 0., 0., 2., 1., 1., 1., 0., 0.])
    
    np.testing.assert_array_almost_equal(predictions, expected)


def test_recursive_predict_output_with_two_window_features():
    """
    Test _recursive_predict output with 2 window features.
    """
    rolling = RollingFeaturesClassification(stats=['proportion'], window_sizes=4)
    rolling_2 = RollingFeaturesClassification(stats=['entropy'], window_sizes=4)
    forecaster = ForecasterRecursiveClassifier(
        LGBMClassifier(verbose=-1), lags=3, window_features=[rolling, rolling_2]
    )
    forecaster.fit(y=y, exog=exog)

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=10, exog=exog_predict)
    )
    predictions = forecaster._recursive_predict(
                      steps              = 10,
                      last_window_values = last_window_values,
                      exog_values        = exog_values
                  )
    
    expected = np.array([1., 1., 0., 0., 2., 1., 1., 1., 0., 0.])
    
    np.testing.assert_array_almost_equal(predictions, expected)



# Unit test transform_dataframe
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from skforecast.utils import transform_dataframe


def test_transform_dataframe_TypeError_when_df_is_not_pandas_DataFrame():
    """
    Test TypeError is raised when df is not a pandas DataFrame.
    """
    df = pd.Series(np.arange(10), name='y')

    err_msg = re.escape(f"`df` argument must be a pandas DataFrame. Got {type(df)}")
    with pytest.raises(TypeError, match = err_msg):
        transform_dataframe(
            df                = df,
            transformer       = None,
            fit               = True,
            inverse_transform = False
        )


def test_transform_dataframe_ValueError_when_transformer_is_ColumnTransformer_and_inverse_transform_is_true():
    """
    Test that transform_dataframe raise ValueError when transformer is ColumnTransformer
    and argument inverse_transform is True.
    """
    df_input = pd.DataFrame({
                   'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4],
                   'col_2': ['a', 'a', 'a', 'a', 'b', 'b', 'b', 'b']
               })
    transformer = ColumnTransformer(
                      [('scale', StandardScaler(), ['col_1']),
                       ('onehot', OneHotEncoder(sparse_output=False), ['col_2'])],
                      remainder = 'passthrough',
                      verbose_feature_names_out = False
                  )

    err_msg = re.escape("`inverse_transform` is not available when using ColumnTransformers.")
    with pytest.raises(ValueError, match = err_msg):
        transform_dataframe(
            df                = df_input,
            transformer       = transformer,
            fit               = True,
            inverse_transform = True
        )


def test_transform_dataframe_when_transformer_is_None():
    """
    Test the output of transform_dataframe when transformer is None.
    """
    df_input = pd.DataFrame({
                   'A': [1.16, -0.28, 0.07, 2.4, 0.25, -0.56, -1.42, 1.26, 1.78, -1.49],
                   'B': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4, 2.7, 59.6]
               })  
    expected = df_input
    transformer = None
    results = transform_dataframe(
                  df                = df_input,
                  transformer       = transformer,
                  fit               = True,
                  inverse_transform = False
              )
    
    pd.testing.assert_frame_equal(results, expected)


def test_transform_dataframe_when_transformer_is_StandardScaler():
    """
    Test the output of transform_dataframe when transformer is StandardScaler.
    """
    df_input = pd.DataFrame({
                   'A': [1.16, -0.28, 0.07, 2.4, 0.25, -0.56, -1.42, 1.26, 1.78, -1.49],
                   'B': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4, 2.7, 59.6]
               })
    expected = pd.DataFrame({
                   'A': [0.67596768, -0.47871021, -0.19805933,  1.67027365, -0.0537246 ,
                         -0.70323091, -1.39283021,  0.75615365,  1.17312067, -1.44896038],
                   'B': [-1.47939551, -0.79158852,  0.6694926 ,  0.54739669,  0.27878567,
                         -0.09971166,  1.76428598,  0.14448017, -1.67474897,  0.64100356]
               })
    transformer = StandardScaler()
    results = transform_dataframe(
                  df                = df_input,
                  transformer       = transformer,
                  fit               = True,
                  inverse_transform = False
              )
    
    pd.testing.assert_frame_equal(results, expected)


def test_transform_dataframe_when_transformer_is_OneHotEncoder():
    """
    Test the output of transform_dataframe when transformer is OneHotEncoder.
    """
    df_input = pd.DataFrame({
                   'col_1': ['a', 'a', 'a', 'a', 'b', 'b', 'b', 'b'],
                   'col_2': ['a', 'a', 'a', 'a', 'b', 'b', 'b', 'b']
               })
    expected = pd.DataFrame({
                   'col_1_a': [1., 1., 1., 1., 0., 0., 0., 0.],
                   'col_1_b': [0., 0., 0., 0., 1., 1., 1., 1.],
                   'col_2_a': [1., 1., 1., 1., 0., 0., 0., 0.],
                   'col_2_b': [0., 0., 0., 0., 1., 1., 1., 1.]
               })
    transformer = OneHotEncoder()
    results = transform_dataframe(
                  df                = df_input,
                  transformer       = transformer,
                  fit               = True,
                  inverse_transform = False
              )
    
    pd.testing.assert_frame_equal(results, expected)


def test_transform_dataframe_when_transformer_is_ColumnTransformer():
    """
    Test the output of transform_dataframe when transformer is ColumnTransformer.
    """
    df_input = pd.DataFrame({
                   'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4],
                   'col_2': ['a', 'a', 'a', 'a', 'b', 'b', 'b', 'b']
               })
    expected = pd.DataFrame({
                   'col_1': [-1.76425513, -1.00989936, 0.59254869, 0.45863938,
                             0.1640389 , -0.25107995, 1.79326881, 0.01673866],
                   'col_2_a': [1., 1., 1., 1., 0., 0., 0., 0.],
                   'col_2_b': [0., 0., 0., 0., 1., 1., 1., 1.]
               })
    transformer = ColumnTransformer(
                      [('scale', StandardScaler(), ['col_1']),
                       ('onehot', OneHotEncoder(), ['col_2'])],
                      remainder = 'passthrough',
                      verbose_feature_names_out = False
                  )
    results = transform_dataframe(
                  df                = df_input,
                  transformer       = transformer,
                  fit               = True,
                  inverse_transform = False
              )
    
    pd.testing.assert_frame_equal(results, expected)



################################################################################
#                            skforecast.exceptions                             #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

"""
The skforecast.exceptions module contains all the custom warnings and error 
classes used across skforecast.
"""
import warnings
import inspect
from functools import wraps
import textwrap
from rich.console import Console
from rich.panel import Panel
from rich.text import Text


def runtime_deprecated(
    replacement: str = None, 
    version: str = None, 
    removal: str = None, 
    category=FutureWarning
) -> object:
    """
    Decorator to mark functions or classes as deprecated.
    Works for both function and class targets, and ensures warnings are visible
    even inside Jupyter notebooks.
    """
    def decorator(obj):
        is_function = inspect.isfunction(obj) or inspect.ismethod(obj)
        is_class = inspect.isclass(obj)

        if not (is_function or is_class):
            raise TypeError("@runtime_deprecated can only be used on functions or classes")

        # ----- Build warning message -----
        name = obj.__name__
        message = f"{name}() is deprecated" if is_function else f"{name} class is deprecated"
        if version:
            message += f" since version {version}"
        if replacement:
            message += f"; use {replacement} instead"
        if removal:
            message += f". It will be removed in version {removal}."
        else:
            message += "."

        def issue_warning():
            """Emit warning in a way that always shows in notebooks."""
            with warnings.catch_warnings():
                warnings.simplefilter("always", category)
                warnings.warn(message, category, stacklevel=3)

        # ----- Case 1: decorating a function -----
        if is_function:
            @wraps(obj)
            def wrapper(*args, **kwargs):
                issue_warning()
                return obj(*args, **kwargs)

            # Add metadata
            wrapper.__deprecated__ = True
            wrapper.__replacement__ = replacement
            wrapper.__version__ = version
            wrapper.__removal__ = removal
            return wrapper

        # ----- Case 2: decorating a class -----
        elif is_class:
            orig_init = getattr(obj, "__init__", None)
            orig_new = getattr(obj, "__new__", None)

            # Only wrap whichever exists (some classes use __new__, others __init__)
            if orig_new and (orig_new is not object.__new__):
                @wraps(orig_new)
                def wrapped_new(cls, *args, **kwargs):
                    issue_warning()
                    return orig_new(cls, *args, **kwargs)
                obj.__new__ = staticmethod(wrapped_new)

            elif orig_init:
                @wraps(orig_init)
                def wrapped_init(self, *args, **kwargs):
                    issue_warning()
                    return orig_init(self, *args, **kwargs)
                obj.__init__ = wrapped_init

            # Add metadata
            obj.__deprecated__ = True
            obj.__replacement__ = replacement
            obj.__version__ = version
            obj.__removal__ = removal

            return obj

    return decorator


class DataTypeWarning(UserWarning):
    """
    Warning used to notify there are dtypes in the exogenous data that are not
    'int', 'float', 'bool' or 'category'. Most machine learning models do not
    accept other data types, therefore the forecaster `fit` and `predict` may fail.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=DataTypeWarning)"
        )
        return self.message + "\n" + extra_message


class DataTransformationWarning(UserWarning):
    """
    Warning used to notify that the output data is in the transformed space.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=DataTransformationWarning)"
        )
        return self.message + "\n" + extra_message


class FeatureOutOfRangeWarning(UserWarning):
    """
    Warning used to notify that a feature is out of the range seen during training.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=FeatureOutOfRangeWarning)"
        )
        return self.message + "\n" + extra_message


class IgnoredArgumentWarning(UserWarning):
    """
    Warning used to notify that an argument is ignored when using a method 
    or a function.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=IgnoredArgumentWarning)"
        )
        return self.message + "\n" + extra_message


class IndexWarning(UserWarning):
    """
    Warning used to notify that the index of the input data is not a
    expected type. 
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=IndexWarning)"
        )
        return self.message + "\n" + extra_message


class InputTypeWarning(UserWarning):
    """
    Warning used to notify that input format is not the most efficient or
    recommended for the forecaster.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=InputTypeWarning)"
        )
        return self.message + "\n" + extra_message


class LongTrainingWarning(UserWarning):
    """
    Warning used to notify that a large number of models will be trained and the
    the process may take a while to run.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=LongTrainingWarning)"
        )
        return self.message + "\n" + extra_message


class MissingExogWarning(UserWarning):
    """
    Warning used to indicate that there are missing exogenous variables in the
    data. Most machine learning models do not accept missing values, so the
    Forecaster's `fit' and `predict' methods may fail.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=MissingExogWarning)"
        )
        return self.message + "\n" + extra_message


class MissingValuesWarning(UserWarning):
    """
    Warning used to indicate that there are missing values in the data. This 
    warning occurs when the input data contains missing values, or the training
    matrix generates missing values. Most machine learning models do not accept
    missing values, so the Forecaster's `fit' and `predict' methods may fail.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=MissingValuesWarning)"
        )
        return self.message + "\n" + extra_message


class OneStepAheadValidationWarning(UserWarning):
    """
    Warning used to notify that the one-step-ahead validation is being used.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=OneStepAheadValidationWarning)"
        )
        return self.message + "\n" + extra_message


class ResidualsUsageWarning(UserWarning):
    """
    Warning used to notify that a residual are not correctly used in the
    probabilistic forecasting process.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=ResidualsUsageWarning)"
        )
        return self.message + "\n" + extra_message


class UnknownLevelWarning(UserWarning):
    """
    Warning used to notify that a level being predicted was not part of the
    training data.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=UnknownLevelWarning)"
        )
        return self.message + "\n" + extra_message


class SaveLoadSkforecastWarning(UserWarning):
    """
    Warning used to notify any issues that may arise when saving or loading
    a forecaster.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=SaveLoadSkforecastWarning)"
        )
        return self.message + "\n" + extra_message


class SkforecastVersionWarning(UserWarning):
    """
    Warning used to notify that the skforecast version installed in the 
    environment differs from the version used to initialize the forecaster.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=SkforecastVersionWarning)"
        )
        return self.message + "\n" + extra_message


warn_skforecast_categories = [
    DataTypeWarning,
    DataTransformationWarning,
    FeatureOutOfRangeWarning,
    IgnoredArgumentWarning,
    IndexWarning,
    InputTypeWarning,
    LongTrainingWarning,
    MissingExogWarning,
    MissingValuesWarning,
    OneStepAheadValidationWarning,
    ResidualsUsageWarning,
    UnknownLevelWarning,
    SaveLoadSkforecastWarning,
    SkforecastVersionWarning
]


def format_warning_handler(
    message: str, 
    category: str, 
    filename: str, 
    lineno: str, 
    file: object = None, 
    line: str = None
) -> None:
    """
    Custom warning handler to format warnings in a box for skforecast custom
    warnings.

    Parameters
    ----------
    message : str
        Warning message.
    category : str
        Warning category.
    filename : str
        Filename where the warning was raised.
    lineno : int
        Line number where the warning was raised.
    file : file, default None
        File where the warning was raised.
    line : str, default None
        Line where the warning was raised.

    Returns
    -------
    None

    """

    if isinstance(message, tuple(warn_skforecast_categories)):
        width = 88
        title = type(message).__name__
        output_text = ["\n"]

        wrapped_message = textwrap.fill(str(message), width=width - 2, expand_tabs=True, replace_whitespace=True)
        title_top_border = f"╭{'─' * ((width - len(title) - 2) // 2)} {title} {'─' * ((width - len(title) - 2) // 2)}╮"
        if len(title) % 2 != 0:
            title_top_border = title_top_border[:-1] + '─' + "╮"
        bottom_border = f"╰{'─' * width}╯"
        output_text.append(title_top_border)

        for line in wrapped_message.split('\n'):
            output_text.append(f"│ {line.ljust(width - 2)} │")

        output_text.append(bottom_border)
        output_text = "\n".join(output_text)
        color = '\033[38;5;208m'
        reset = '\033[0m'
        output_text = f"{color}{output_text}{reset}"
        print(output_text)
    else:
        # Fallback to default Python warning formatting
        warnings._original_showwarning(message, category, filename, lineno, file, line)


def rich_warning_handler(
    message: str, 
    category: str, 
    filename: str, 
    lineno: str, 
    file: object = None, 
    line: str = None
) -> None:
    """
    Custom handler for warnings that uses rich to display formatted panels.

    Parameters
    ----------
    message : str
        Warning message.
    category : str
        Warning category.
    filename : str
        Filename where the warning was raised.
    lineno : int
        Line number where the warning was raised.
    file : file, default None
        File where the warning was raised.
    line : str, default None
        Line where the warning was raised.

    Returns
    -------
    None

    """
    
    if isinstance(message, tuple(warn_skforecast_categories)):
        console = Console()

        category_name = category.__name__
        text = (
            f"{message.message}\n\n"
            f"Category : skforecast.exceptions.{category_name}\n"
            f"Location : {filename}:{lineno}\n"
            f"Suppress : warnings.simplefilter('ignore', category={category_name})"
        )

        panel = Panel(
            Text(text, justify="left"),
            title        = category_name,
            title_align  = "center",
            border_style = "color(214)",
            width        = 88,
        )
        
        console.print(panel)
    else:
        # Fallback to default Python warning formatting
        warnings._original_showwarning(message, category, filename, lineno, file, line)


def set_warnings_style(style: str = 'skforecast') -> None:
    """
    Set the warning handler based on the provided style.

    Parameters
    ----------
    style : str, default='skforecast'
        The style of the warning handler. Either 'skforecast' or 'default'.
    
    Returns
    -------
    None

    """
    if style == "skforecast":
        if not hasattr(warnings, "_original_showwarning"):
            warnings._original_showwarning = warnings.showwarning
        warnings.showwarning = rich_warning_handler
    else:
        warnings.showwarning = warnings._original_showwarning


set_warnings_style(style='skforecast')



# Unit test get_feature_importances ForecasterRecursiveMultiSeries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from sklearn.neural_network import MLPRegressor
from sklearn.ensemble import RandomForestRegressor
from lightgbm import LGBMRegressor

from skforecast.preprocessing import RollingFeatures, reshape_series_wide_to_long
from ....recursive import ForecasterRecursiveMultiSeries

# Fixtures
from .fixtures_forecaster_recursive_multiseries import (
    series_dict_nans,
    exog_dict_nans
)
series = pd.DataFrame(
    {'1': np.arange(10), '2': np.arange(10)},
    index = pd.date_range(start='2000-01-01', periods=10, freq='D')
)
series = reshape_series_wide_to_long(series)
exog = pd.Series(np.arange(10, 20), name='exog')
exog.index = pd.date_range(start='2000-01-01', periods=len(exog), freq='D')


def test_NotFittedError_is_raised_when_forecaster_is_not_fitted():
    """
    Test NotFittedError is raised when calling get_feature_importances() and 
    forecaster is not fitted.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator = LinearRegression(),
                     lags      = 3,
                 )

    err_msg = re.escape(
        "This forecaster is not fitted yet. Call `fit` with appropriate "
        "arguments before using `get_feature_importances()`."
    )
    with pytest.raises(NotFittedError, match = err_msg):         
        forecaster.get_feature_importances()


def test_output_get_feature_importances_when_estimator_is_RandomForest():
    """
    Test output of get_feature_importances when estimator is RandomForestRegressor with lags=3
    and it is trained with series pandas DataFrame.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator = RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123),
                     lags      = 3,
                     encoding  = 'onehot'
                 )
    forecaster.fit(series=series)
    results = forecaster.get_feature_importances(sort_importance=False)

    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2'],
                   'importance': np.array([0.9446366782006932, 0.0, 0.05536332179930687, 0.0, 0.0])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_estimator_is_RandomForest_with_exog():
    """
    Test output of get_feature_importances when estimator is RandomForestRegressor with lags=3
    and it is trained with series pandas DataFrame and a exogenous variable.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator = RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123),
                     lags      = 3,
                     encoding  = 'onehot'
                 )
    forecaster.fit(series=series, exog=exog)
    results = forecaster.get_feature_importances()

    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2', 'exog'],
                   'importance': np.array([0.73269896, 0., 0.21193772, 0., 0., 0.05536332])
               }).sort_values(by='importance', ascending=False)

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_estimator_is_LinearRegression():
    """
    Test output of get_feature_importances when estimator is LinearRegression with lags=3
    and it is trained with series pandas DataFrame.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator = LinearRegression(),
                     lags      = 3,
                     encoding  = 'onehot'
                 )
    forecaster.fit(series=series)
    results = forecaster.get_feature_importances(sort_importance=False)

    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2'],
                   'importance': np.array([3.33333333e-01, 3.33333333e-01, 3.33333333e-01, 
                                           -1.48164367e-16, 1.48164367e-16])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_with_exog():
    """
    Test output of get_feature_importances when estimator is LinearRegression with lags=3
    and it is trained with series pandas DataFrame and a exogenous variable.
    """

    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     encoding           = 'onehot',
                     transformer_series = None
                 )
    forecaster.fit(series=series, exog=exog)
    results = forecaster.get_feature_importances(sort_importance=False)

    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2', 'exog'],
                   'importance': np.array([2.50000000e-01,  2.50000000e-01,  2.50000000e-01,  
                                           -2.97120907e-17,  2.97120907e-17, 2.50000000e-01])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_and_UserWarning_get_feature_importances_when_estimator_no_attributes():
    """
    Test output of get_feature_importances when estimator is MLPRegressor with lags=3
    and it is trained with series pandas DataFrame. Since MLPRegressor hasn't attributes
    `feature_importances_` or `coef_, results = None and a UserWarning is issues.
    """
    series_2 = {
        '1': pd.Series(np.arange(5)), 
        '2': pd.Series(np.arange(5))
    }
    forecaster = ForecasterRecursiveMultiSeries(
        MLPRegressor(solver = 'lbfgs', max_iter= 50, random_state=123), lags=3
    )
    forecaster.fit(series=series_2)

    estimator = forecaster.estimator
    expected = None

    warn_msg = re.escape(
        f"Impossible to access feature importances for estimator of type "
        f"{type(estimator)}. This method is only valid when the "
        f"estimator stores internally the feature importances in the "
        f"attribute `feature_importances_` or `coef_`."
    )
    with pytest.warns(UserWarning, match = warn_msg):
        results = forecaster.get_feature_importances()
        assert results is expected


def test_output_get_feature_importances_when_pipeline_LinearRegression():
    """
    Test output of get_feature_importances when estimator is pipeline,
    (StandardScaler() + LinearRegression with lags=3),
    it is trained with series pandas DataFrame.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = make_pipeline(StandardScaler(), LinearRegression()),
                     lags               = 3,
                     encoding           = 'onehot',
                     transformer_series = None
                 )
    forecaster.fit(series=series)
    results = forecaster.get_feature_importances(sort_importance=False)

    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2'],
                   'importance': np.array([6.66666667e-01, 6.66666667e-01, 6.66666667e-01, 
                                           -7.40821837e-17, 7.40821837e-17])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_pipeline_RandomForestRegressor():
    """
    Test output of get_feature_importances when estimator is pipeline,
    (StandardScaler() + RandomForestRegressor with lags=3),
    it is trained with series pandas DataFrame.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator = make_pipeline(
                                     StandardScaler(), 
                                     RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123)
                                 ),
                     lags               = 3,
                     encoding           = "onehot",
                     transformer_series = None,
                 )
    forecaster.fit(series=series)
    results = forecaster.get_feature_importances(sort_importance=False)

    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2'],
                   'importance': np.array([0.9446366782006932, 0.0, 0.05536332179930687, 0.0, 0.0])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_estimator_is_RandomForest_with_exog_ordinal():
    """
    Test output of get_feature_importances when estimator is RandomForestRegressor with lags=3
    and it is trained with series pandas DataFrame and a exogenous variable 
    with encoding='ordinal'.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator = RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123),
                     lags      = 3,
                     encoding  = 'ordinal'
                 )
    forecaster.fit(series=series, exog=exog)
    results = forecaster.get_feature_importances()

    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '_level_skforecast', 'exog'],
                   'importance': np.array([0.944637, 0., 0.055363, 0., 0.])
               }).sort_values(by='importance', ascending=False)

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_estimator_is_RandomForest_with_exog_ordinal_category():
    """
    Test output of get_feature_importances when estimator is RandomForestRegressor with lags=3
    and it is trained with series pandas DataFrame and a exogenous variable 
    with encoding='ordinal_category'.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator = RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123),
                     lags      = 3,
                     encoding  = 'ordinal_category'
                 )
    forecaster.fit(series=series, exog=exog)
    results = forecaster.get_feature_importances()

    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '_level_skforecast', 'exog'],
                   'importance': np.array([0.944637, 0., 0.055363, 0., 0.])
               }).sort_values(by='importance', ascending=False)

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_estimator_is_RandomForest_with_exog_None_encoding():
    """
    Test output of get_feature_importances when estimator is RandomForestRegressor with lags=3
    and it is trained with series pandas DataFrame and a exogenous variable 
    with encoding=None.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator = RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123),
                     lags      = 3,
                     encoding  = None
                 )
    forecaster.fit(series=series, exog=exog)
    results = forecaster.get_feature_importances()

    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', 'exog'],
                   'importance': np.array([0.732699, 0.211938, 0.055363, 0.])
               }).sort_values(by='importance', ascending=False)

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_window_features():
    """
    Test output of get_feature_importances when estimator is LGMBRegressor with 
    lags=3 and window features.
    """    
    rolling = RollingFeatures(stats=['mean', 'sum'], window_sizes=[3, 5])
    forecaster = ForecasterRecursiveMultiSeries(
        LGBMRegressor(verbose=-1, random_state=123), lags=3, window_features=rolling
    )
    forecaster.fit(series=series_dict_nans, exog=exog_dict_nans)

    results = forecaster.get_feature_importances(sort_importance=False)
    results = results.astype({'importance': float})
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', 'roll_mean_3', 'roll_sum_5', '_level_skforecast', 
                               'sin_day_of_week', 'cos_day_of_week', 'air_temperature', 'wind_speed'],
                   'importance': np.array([480.0, 361.0, 323.0, 251.0, 430.0, 
                                           41.0, 175.0, 105.0, 468.0, 366.0])
               }).astype({'importance': float})
    
    pd.testing.assert_frame_equal(results, expected)



# Unit test prepare_levels_multiseries
# ==============================================================================
import pytest
from skforecast.utils import prepare_levels_multiseries


@pytest.mark.parametrize("levels, expected_levels, expected_type", 
                         [(None, ['l1', 'l2', 'l3'], False),
                          ('l1', ['l1'], False),
                          (['l1'], ['l1'], True)], 
                         ids=lambda levels: f'levels: {levels}')
def test_output_prepare_levels_multiseries(levels, expected_levels, expected_type):
    """
    Test output prepare_levels_multiseries for different inputs.
    """
    X_train_series_names_in_ = ['l1', 'l2', 'l3']
    levels, input_levels_is_list = prepare_levels_multiseries(
        X_train_series_names_in_=X_train_series_names_in_, levels=levels
    )
    
    assert levels == expected_levels
    assert input_levels_is_list is expected_type



# Unit test reshape_series_exog_dict_to_long
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from ...preprocessing import reshape_series_exog_dict_to_long


def test_reshape_series_exog_dict_to_long_raises_ValueError_when_both_None():
    """
    Test that ValueError is raised when both series and exog are None.
    """
    with pytest.raises(ValueError, match="Both `series` and `exog` cannot be None."):
        reshape_series_exog_dict_to_long(series=None, exog=None)


def test_reshape_series_exog_dict_to_long_raises_TypeError_when_series_not_dict():
    """
    Test that TypeError is raised when series is not a dictionary.
    """
    with pytest.raises(TypeError, match="`series` must be a dictionary"):
        reshape_series_exog_dict_to_long(series=[1, 2, 3], exog=None)


def test_reshape_series_exog_dict_to_long_raises_TypeError_when_series_value_not_Series():
    """
    Test that TypeError is raised when a value in series dict is not a pandas Series.
    """
    series = {
        'series_1': pd.Series([1, 2, 3]),
        'series_2': [4, 5, 6]  # Not a Series
    }
    
    error_msg = re.escape("`series['series_2']` must be a pandas Series.")
    with pytest.raises(TypeError, match=error_msg):
        reshape_series_exog_dict_to_long(series=series, exog=None)


def test_reshape_series_exog_dict_to_long_raises_TypeError_when_exog_not_dict():
    """
    Test that TypeError is raised when exog is not a dictionary.
    """
    with pytest.raises(TypeError, match="`exog` must be a dictionary"):
        reshape_series_exog_dict_to_long(series=None, exog=[1, 2, 3])


def test_reshape_series_exog_dict_to_long_raises_TypeError_when_exog_value_invalid():
    """
    Test that TypeError is raised when a value in exog dict is not a Series or DataFrame.
    """
    exog = {
        'series_1': pd.Series([1, 2, 3]),
        'series_2': [4, 5, 6]  # Not a Series or DataFrame
    }

    error_msg = re.escape(
        "`exog['series_2']` must be a pandas Series or a pandas DataFrame."
    )
    with pytest.raises(TypeError, match=error_msg):
        reshape_series_exog_dict_to_long(series=None, exog=exog)


def test_TypeError_series_exog_dict_to_long_different_index_type():
    """
    Test TypeError is raised when series and exog have different index types.
    """
    series = {
        'series_1': pd.Series([1, 2, 3], index=pd.date_range('2020-01-01', periods=3)),
        'series_2': pd.Series([4, 5, 6], index=pd.date_range('2020-01-01', periods=3))
    }
    exog = {
        'series_1': pd.Series([1, 2, 3], name='exog_1'),
        'series_2': pd.Series([4, 5, 6], name='exog_2')
    }

    series_idx_type = "<class 'pandas.core.indexes.datetimes.DatetimeIndex'>"
    exog_idx_type = "<class 'pandas.core.indexes.base.Index'>"

    error_msg = re.escape(
        f"Index type mismatch: series has index of type "
        f"{series_idx_type}, but `exog` has {exog_idx_type}. "
        f"Ensure all indices are compatible."
    )
    with pytest.raises(TypeError, match=error_msg):
        reshape_series_exog_dict_to_long(series=series, exog=exog)


def test_ValueError_when_series_col_name_in_exog_columns():
    """
    Test ValueError is raised when `series_col_name` is already in exog columns.
    """
    series = {
        'series_1': pd.Series([1, 2, 3], index=pd.date_range('2020-01-01', periods=3)),
        'series_2': pd.Series([4, 5, 6], index=pd.date_range('2020-01-01', periods=3))
    }
    exog = {
        'series_1': pd.DataFrame(
            {'exog1': [1, 2, 3], 'exog2': [7, 8, 9]},
            index=pd.date_range('2020-01-01', periods=3)
        ),
        'series_2': pd.DataFrame(
            {'exog1': [4, 5, 6], 'exog2': [10, 11, 12]},
            index=pd.date_range('2020-01-01', periods=3)
        )
    }

    error_msg = re.escape(
        "Column name conflict: 'exog1' already exists in `exog`. "
        "Please choose a different `series_col_name` value."
    )
    with pytest.raises(ValueError, match=error_msg):
        reshape_series_exog_dict_to_long(
            series=series, exog=exog, series_col_name='exog1'
        )


def test_reshape_series_exog_dict_to_long_only_series():
    """
    Test reshaping when only series dict is provided.
    """
    series = {
        'series_1': pd.Series([1, 2, 3], index=pd.date_range('2020-01-01', periods=3)),
        'series_2': pd.Series([4, 5, 6], index=pd.date_range('2020-01-01', periods=3))
    }
    
    result = reshape_series_exog_dict_to_long(series=series, exog=None)
    expected_result = pd.DataFrame(
        data={"series_value": [1, 2, 3, 4, 5, 6]},
        index=pd.MultiIndex.from_product(
            [["series_1", "series_2"], pd.date_range("2020-01-01", periods=3)],
            names=["series_id", "datetime"],
        ),
    )

    pd.testing.assert_frame_equal(result, expected_result)


def test_reshape_series_exog_dict_to_long_only_exog_Series():
    """
    Test reshaping when only exog dict with Series is provided.
    """
    exog = {
        'series_1': pd.Series([1, 2, 3], index=pd.date_range('2020-01-01', periods=3), name='exog1'),
        'series_2': pd.Series([4, 5, 6], index=pd.date_range('2020-01-01', periods=3), name='exog1')
    }
    
    result = reshape_series_exog_dict_to_long(series=None, exog=exog)
    expected_result = pd.DataFrame(
        data={"exog_value": [1, 2, 3, 4, 5, 6]},
        index=pd.MultiIndex.from_product(
            [["series_1", "series_2"], pd.date_range("2020-01-01", periods=3)],
            names=["series_id", "datetime"],
        ),
    )

    pd.testing.assert_frame_equal(result, expected_result)


def test_reshape_series_exog_dict_to_long_only_exog_DataFrame():
    """
    Test reshaping when only exog dict with DataFrames is provided.
    """
    exog = {
        'series_1': pd.DataFrame(
            {'exog1': [1, 2, 3], 'exog2': [7, 8, 9]},
            index=pd.date_range('2020-01-01', periods=3)
        ),
        'series_2': pd.DataFrame(
            {'exog1': [4, 5, 6], 'exog2': [10, 11, 12]},
            index=pd.date_range('2020-01-01', periods=3)
        )
    }
    
    result = reshape_series_exog_dict_to_long(series=None, exog=exog)
    expected_result = pd.DataFrame(
        data={"exog1": [1, 2, 3, 4, 5, 6], "exog2": [7, 8, 9, 10, 11, 12]},
        index=pd.MultiIndex.from_product(
            [["series_1", "series_2"], pd.date_range("2020-01-01", periods=3)],
            names=["series_id", "datetime"],
        ),
    )

    pd.testing.assert_frame_equal(result, expected_result)


def test_reshape_series_exog_dict_to_long_series_and_exog():
    """
    Test reshaping when both series and exog dicts are provided.
    """
    series_dict = {
        "series_1": pd.Series(
            data=np.arange(3), index=pd.date_range(start="2020-01-01", periods=3, freq="D")
        ),
        "series_2": pd.Series(
            data=np.arange(3, 6),
            index=pd.date_range(start="2020-01-01", periods=3, freq="D"),
        ),
    }

    exog_dict = {
        "series_1": pd.DataFrame(
            data={
                "exog_1": np.arange(3),
                "exog_2": np.arange(10, 13),
                "exog_3": np.arange(20, 23),
            },
            index=pd.date_range(start="2020-01-01", periods=3, freq="D"),
        ),
        "series_2": pd.DataFrame(
            data={"exog_1": np.arange(3, 6), "exog_2": np.arange(13, 16)},
            index=pd.date_range(start="2020-01-01", periods=3, freq="D"),
        ),
    }
    result = reshape_series_exog_dict_to_long(
        series=series_dict,
        exog=exog_dict
    )
    expected_result = pd.DataFrame(
        data={
            "series_value": [0, 1, 2, 3, 4, 5],
            "exog_1": [0, 1, 2, 3, 4, 5],
            "exog_2": [10, 11, 12, 13, 14, 15],
            "exog_3": [20, 21, 22, np.nan, np.nan, np.nan],
        },
        index=pd.MultiIndex.from_product(
            [["series_1", "series_2"], pd.date_range("2020-01-01", periods=3)],
            names=["series_id", "datetime"],
        ),
    )

    pd.testing.assert_frame_equal(result, expected_result)



# Unit test predict
# ==============================================================================
import re
import pytest
import pandas as pd
import warnings
from skforecast.drift_detection import RangeDriftDetector
from skforecast.exceptions import (
    FeatureOutOfRangeWarning,
    UnknownLevelWarning,
    MissingExogWarning
)


def test_predict_not_fitted():
    """
    Test predict raises RuntimeError when detector is not fitted.
    """
    detector = RangeDriftDetector()
    with pytest.raises(RuntimeError, match="Model is not fitted yet."):
        detector.predict()


def test_predict_invalid_last_window_type():
    """
    Test predict raises TypeError for invalid last_window type.
    """
    detector = RangeDriftDetector()
    detector.is_fitted = True

    msg = re.escape(
        "`last_window` must be a pandas Series, DataFrame, dict or None."
    )
    with pytest.raises(TypeError, match=msg):
        detector.predict(last_window="invalid")


def test_predict_invalid_exog_type():
    """
    Test predict raises TypeError for invalid exog type.
    """
    detector = RangeDriftDetector()
    detector.is_fitted = True 

    msg = re.escape(
        "`exog` must be a pandas Series, DataFrame, dict or None."
    )
    with pytest.raises(TypeError, match=msg):
        detector.predict(exog="invalid")


def test_predict_no_out_of_range():
    """
    Test predict with data within training ranges.
    """
    
    series = pd.Series([1, 2, 3, 4, 5], name='series_1')
    exog = pd.DataFrame({'exog1': [10, 20, 30]})
    detector = RangeDriftDetector()
    detector.fit(series=series, exog=exog)

    last_window = pd.Series([2, 3, 4], name='series_1')
    exog_pred = pd.DataFrame({'exog1': [15, 25]})

    flag, out_series, out_exog = detector.predict(
        last_window=last_window, exog=exog_pred, verbose=False
    )

    assert flag is False
    assert out_series == []
    assert out_exog == []


def test_predict_out_of_range_series_pandas_series():
    """
    Test predict with series values out of training range.
    Series is a single pd.Series.
    """
    series = pd.Series([1, 2, 3, 4, 5], name='series_1')
    detector = RangeDriftDetector()
    detector.fit(series=series)

    last_window = pd.Series([0, 6], name='series_1')

    with pytest.warns(FeatureOutOfRangeWarning):
        flag, out_series, out_exog = detector.predict(
            last_window=last_window, verbose=False
        )

    assert flag is True
    assert out_series == ['series_1']
    assert out_exog == []


def test_predict_out_of_range_series_pandas_dataframe():
    """
    Test predict with series values out of training range.
    Series is a pandas dataframe with multiple columns.
    """
    series = pd.DataFrame({
        'series_1': [1, 2, 3, 4, 5],
        'series_2': [10, 20, 30, 40, 50],
        'series_3': [100, 200, 300, 400, 500]
    })
    detector = RangeDriftDetector()
    detector.fit(series=series)

    last_window = pd.DataFrame({
        'series_1': [0, 6],  # out of range
        'series_2': [15, 25],  # in range
        'series_3': [600, 250]  # out of range
    })

    with pytest.warns(FeatureOutOfRangeWarning):
        flag, out_series, out_exog = detector.predict(
            last_window=last_window, verbose=False
        )

    assert flag is True
    assert out_series == ['series_1', 'series_3']
    assert out_exog == []


def test_predict_out_of_range_series_pandas_dataframe_multiindex():
    """
    Test predict with series values out of training range.
    Series is a pandas dataframe multiindex.
    """
    index = pd.MultiIndex.from_product(
                [['series_1', 'series_2', 'series_3'], range(3)],
                names=['series', 'time']
            )
    series = pd.DataFrame({
        'value': [1, 2, 3, 10, 20, 30, 100, 200, 300]
    }, index=index) 

    detector = RangeDriftDetector()
    detector.fit(series=series)

    last_window = pd.DataFrame({
        'series_1': [0, 6],  # out of range
        'series_2': [15, 25],  # in range
        'series_3': [600, 700]  # out of range
    })

    with pytest.warns(FeatureOutOfRangeWarning):
        flag, out_series, out_exog = detector.predict(
            last_window=last_window, verbose=False
        )

    assert flag is True
    assert out_series == ['series_1', 'series_3']
    assert out_exog == []


def test_predict_out_of_range_exog_pandas_series(): 
    """
    Test predict with exogenous values out of training range.
    """
    series = pd.Series([1, 2, 3], name='series_1')
    exog = pd.DataFrame({'exog1': [10, 20, 30]}, index=[0, 1, 2])
    detector = RangeDriftDetector()
    detector.fit(series=series, exog=exog)

    exog_pred = pd.DataFrame({'exog1': [5, 35]}, index=[0, 1])

    with pytest.warns(FeatureOutOfRangeWarning):
        flag, out_series, out_exog = detector.predict(
            exog=exog_pred, verbose=False
        )

    assert flag is True
    assert out_series == []
    assert out_exog == ['exog1']


def test_predict_out_of_range_exog_pandas_dataframe():
    """
    Test predict with exogenous values out of training range when exog is a
    DataFrame with multiple columns.
    """
    series = pd.Series([1, 2, 3], name='series_1')
    exog = pd.DataFrame({'exog1': [10, 20, 30], 'exog2': [100, 200, 300]}, index=[0, 1, 2])
    detector = RangeDriftDetector()
    detector.fit(series=series, exog=exog)

    exog_pred = pd.DataFrame({
        'exog1': [5, 35],  # out of range
        'exog2': [95, 205]},  # out of range
        index=[0, 1]
    )

    with pytest.warns(FeatureOutOfRangeWarning):
        flag, out_series, out_exog = detector.predict(
            exog=exog_pred, verbose=False
        )

    assert flag is True
    assert out_series == []
    assert out_exog == ['exog1', 'exog2']


def test_predict_out_of_range_exog_pandas_dataframe_multiindex():
    """
    Test predict with exogenous values out of training range when exog is a
    DataFrame with MultiIndex.
    """
    index = pd.MultiIndex.from_product(
                [['series_1', 'series_2', 'series_3'], range(3)],
                names=['series_id', 'time']
            )
    series = pd.DataFrame(
        {"values": [1, 2, 3, 10, 20, 30, 100, 200, 300]}, index=index
    )
    exog = pd.DataFrame(
        {'exog1': [10, 20, 30, 100, 200, 300, 1000, 2000, 3000],
        'exog2': [-10, -20, -30, -100, -200, -300, -1000, -2000, -3000]},
        index=index
    )
    detector = RangeDriftDetector()
    detector.fit(series=series, exog=exog)

    index_pred = pd.MultiIndex.from_product(
        [['series_1', 'series_2', 'series_3'], range(2)],
        names=['series_id', 'time']
    )
    # Series 1, exog 1: in range
    # Series 1, exog 2: out of range
    # Series 2, exog 1: out of range
    # Series 2, exog 2: in range
    # Series 3, exog 1: out of range
    # Series 3, exog 2: out of range
    exog_pred = pd.DataFrame(
        {'exog1': [15, 25, 400, 500, 4000, 5000],
         'exog2': [-15, -35, -200, -200, -4000, -5000]},
        index=index_pred
    )

    with pytest.warns(FeatureOutOfRangeWarning):
        flag, out_series, out_exog = detector.predict(
            exog=exog_pred, verbose=False
        )

    expected_out_exog = {
        'series_1': ['exog2'], 
        'series_2': ['exog1'], 
        'series_3': ['exog1', 'exog2']
    }

    assert flag is True
    assert out_series == []
    assert set(out_exog.keys()) == set(expected_out_exog.keys())
    for k in out_exog.keys():
        assert out_exog[k] == expected_out_exog[k]


def test_predict_unknown_series():
    """
    Test predict with series not seen during training.
    """

    series = pd.Series([1, 2, 3], name='series_1')
    detector = RangeDriftDetector()
    detector.fit(series=series)
    last_window = pd.Series([1, 2], name='unknown_series')

    msg = "'unknown_series' was not seen during training. Its range is unknown."
    with pytest.warns(UnknownLevelWarning, match=msg):
        flag, out_series, out_exog = detector.predict(
            last_window=last_window, verbose=False
        )

    assert flag is False
    assert out_series == []
    assert out_exog == []


def test_predict_unknown_exog():
    """
    Test predict with exogenous variable not seen during training.
    """
    series = pd.Series([1, 2, 3], name='series_1')
    exog = pd.DataFrame({'exog1': [10, 20, 30]}, index=[0, 1, 2])
    detector = RangeDriftDetector()
    detector.fit(series=series, exog=exog)

    exog_pred = pd.DataFrame({'unknown_exog': [1, 2]}, index=[0, 1])
    msg = "'unknown_exog' was not seen during training. Its range is unknown."
    with pytest.warns(MissingExogWarning, match=msg):
        flag, out_series, out_exog = detector.predict(
            exog=exog_pred, verbose=False
        )

    assert flag is False
    assert out_series == []
    assert out_exog == []


def test_predict_suppress_warnings():
    """
    Test predict with suppress_warnings=True suppresses warnings.
    """
    # Fit detector
    series = pd.Series([1, 2, 3], name='series_1')
    detector = RangeDriftDetector()
    detector.fit(series=series)

    # Test data out of range
    last_window = pd.Series([0], name='series_1')

    with warnings.catch_warnings():
        warnings.simplefilter("error")  # Turn warnings into errors
        flag, out_series, out_exog = detector.predict(
            last_window=last_window, suppress_warnings=True, verbose=False
        )


def test_predict_with_dict_inputs_series_and_exog():
    """
    Test predict with dict inputs for both series and exog.
    """
    # Fit detector
    series = {
        'series_1': pd.Series([1, 2, 3], name='series_1'),
        'series_2': pd.Series([10, 20, 30], name='series_2')
    }
    exog = {
        'series_1': pd.DataFrame({'exog_1': [10, 20], 'exog_2': [-10, -20]}, index=[0, 1]),
        'series_2': pd.DataFrame({'exog_1': [100, 200], 'exog_2': [-100, -200]}, index=[0, 1])
    }
    detector = RangeDriftDetector()
    detector.fit(series=series, exog=exog)

    last_window = {
        'series_1': pd.Series([-10, 0, -5], name='series_1'),  # out of range
        'series_2': pd.Series([15, 16, 17], name='series_2')   # in range
    }
    exog_pred = {
        'series_1': pd.DataFrame({
            'exog_1': [11, 12, 13],  # in range
            'exog_2': [-5, -1, -0]   # out of range
        }, index=[0, 1, 2]),
        'series_2': pd.DataFrame({
            'exog_1': [100, 110, 120],  # in range
            'exog_2': [-50, -60, -70]   # out of range
        }, index=[0, 1, 2])
    }

    with pytest.warns(FeatureOutOfRangeWarning):
        flag_out_of_range, out_of_range_series, out_of_range_exog = detector.predict(
            last_window=last_window, exog=exog_pred, verbose=True, suppress_warnings=False
        )

    expected_out_exog = {
        'series_1': ['exog_2'], 
        'series_2': ['exog_2'], 
    }

    assert flag_out_of_range is True
    assert out_of_range_series == ['series_1']
    assert set(out_of_range_exog.keys()) == set(expected_out_exog.keys())
    for k in out_of_range_exog.keys():
        assert out_of_range_exog[k] == expected_out_exog[k]



# Unit test create_sample_weights ForecasterRecursiveMultiSeries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from ....recursive import ForecasterRecursiveMultiSeries
from ....preprocessing import reshape_series_wide_to_long


def custom_weights(index):  # pragma: no cover
    """
    Return 0 if index is between '2022-01-08' and '2022-01-10', 1 otherwise.
    """
    weights = np.where((index >= "2022-01-08") & (index <= "2022-01-10"), 0, 1)

    return weights


def custom_weights_2(index):  # pragma: no cover
    """
    Return 2 if index is between '2022-01-11' and '2022-01-13', 3 otherwise.
    """
    weights = np.where((index >= "2022-01-11") & (index <= "2022-01-13"), 2, 3)

    return weights


def custom_weights_nan(index):  # pragma: no cover
    """
    Return np.nan if index is between '2022-01-08' and '2022-01-10', 1 otherwise.
    """
    weights = np.where((index >= "2022-01-08") & (index <= "2022-01-10"), np.nan, 1)

    return weights


def custom_weights_negative(index):  # pragma: no cover
    """
    Return -1 if index is between '2022-01-08' and '2022-01-10', 1 otherwise.
    """
    weights = np.where((index >= "2022-01-08") & (index <= "2022-01-10"), -1, 1)

    return weights


series = pd.DataFrame(
    data=np.array(
        [
            [0.12362923, 0.51328688],
            [0.65138268, 0.11599708],
            [0.58142898, 0.72350895],
            [0.72969992, 0.10305721],
            [0.97790567, 0.20581485],
            [0.56924731, 0.41262027],
            [0.85369084, 0.82107767],
            [0.75425194, 0.0107816],
            [0.08167939, 0.94951918],
            [0.00249297, 0.55583355],
        ]
    ),
    columns=["series_1", "series_2"],
    index=pd.date_range(start="2022-01-04", periods=10, freq="D")
)
series = reshape_series_wide_to_long(series)

X_train_onehot = pd.DataFrame(
    data=np.array(
        [
            [0.58142898, 0.65138268, 0.12362923, 1.0, 0.0],
            [0.72969992, 0.58142898, 0.65138268, 1.0, 0.0],
            [0.97790567, 0.72969992, 0.58142898, 1.0, 0.0],
            [0.56924731, 0.97790567, 0.72969992, 1.0, 0.0],
            [0.85369084, 0.56924731, 0.97790567, 1.0, 0.0],
            [0.75425194, 0.85369084, 0.56924731, 1.0, 0.0],
            [0.08167939, 0.75425194, 0.85369084, 1.0, 0.0],
            [0.72350895, 0.11599708, 0.51328688, 0.0, 1.0],
            [0.10305721, 0.72350895, 0.11599708, 0.0, 1.0],
            [0.20581485, 0.10305721, 0.72350895, 0.0, 1.0],
            [0.41262027, 0.20581485, 0.10305721, 0.0, 1.0],
            [0.82107767, 0.41262027, 0.20581485, 0.0, 1.0],
            [0.0107816, 0.82107767, 0.41262027, 0.0, 1.0],
            [0.94951918, 0.0107816, 0.82107767, 0.0, 1.0],
        ]
    ),
    columns=["lag_1", "lag_2", "lag_3", "series_1", "series_2"],
    index=pd.DatetimeIndex(
        [
            "2022-01-07",
            "2022-01-08",
            "2022-01-09",
            "2022-01-10",
            "2022-01-11",
            "2022-01-12",
            "2022-01-13",
            "2022-01-07",
            "2022-01-08",
            "2022-01-09",
            "2022-01-10",
            "2022-01-11",
            "2022-01-12",
            "2022-01-13",
        ],
        dtype="datetime64[ns]",
        freq=None,
    ),
)

X_train_ordinal = pd.DataFrame(
    data=np.array(
        [
            [0.58142898, 0.65138268, 0.12362923, 0.0],
            [0.72969992, 0.58142898, 0.65138268, 0.0],
            [0.97790567, 0.72969992, 0.58142898, 0.0],
            [0.56924731, 0.97790567, 0.72969992, 0.0],
            [0.85369084, 0.56924731, 0.97790567, 0.0],
            [0.75425194, 0.85369084, 0.56924731, 0.0],
            [0.08167939, 0.75425194, 0.85369084, 0.0],
            [0.72350895, 0.11599708, 0.51328688, 1.0],
            [0.10305721, 0.72350895, 0.11599708, 1.0],
            [0.20581485, 0.10305721, 0.72350895, 1.0],
            [0.41262027, 0.20581485, 0.10305721, 1.0],
            [0.82107767, 0.41262027, 0.20581485, 1.0],
            [0.0107816, 0.82107767, 0.41262027, 1.0],
            [0.94951918, 0.0107816, 0.82107767, 1.0],
        ]
    ),
    columns=["lag_1", "lag_2", "lag_3", "_level_skforecast"],
    index=pd.DatetimeIndex(
        [
            "2022-01-07",
            "2022-01-08",
            "2022-01-09",
            "2022-01-10",
            "2022-01-11",
            "2022-01-12",
            "2022-01-13",
            "2022-01-07",
            "2022-01-08",
            "2022-01-09",
            "2022-01-10",
            "2022-01-11",
            "2022-01-12",
            "2022-01-13",
        ],
        dtype="datetime64[ns]",
        freq=None,
    ),
)

X_train_ordinal_category = X_train_ordinal.copy()
X_train_ordinal_category["_level_skforecast"] = X_train_ordinal_category[
    "_level_skforecast"
].astype("category")

X_train_onehot_diferent_length = pd.DataFrame(
    data=np.array(
        [
            [0.58142898, 0.65138268, 0.12362923, 1.0, 0.0],
            [0.72969992, 0.58142898, 0.65138268, 1.0, 0.0],
            [0.97790567, 0.72969992, 0.58142898, 1.0, 0.0],
            [0.56924731, 0.97790567, 0.72969992, 1.0, 0.0],
            [0.85369084, 0.56924731, 0.97790567, 1.0, 0.0],
            [0.75425194, 0.85369084, 0.56924731, 1.0, 0.0],
            [0.08167939, 0.75425194, 0.85369084, 1.0, 0.0],
            [0.41262027, 0.20581485, 0.10305721, 0.0, 1.0],
            [0.82107767, 0.41262027, 0.20581485, 0.0, 1.0],
            [0.0107816, 0.82107767, 0.41262027, 0.0, 1.0],
            [0.94951918, 0.0107816, 0.82107767, 0.0, 1.0],
        ]
    ),
    columns=["lag_1", "lag_2", "lag_3", "series_1", "series_2"],
    index=pd.DatetimeIndex(
        [
            "2022-01-07",
            "2022-01-08",
            "2022-01-09",
            "2022-01-10",
            "2022-01-11",
            "2022-01-12",
            "2022-01-13",
            "2022-01-10",
            "2022-01-11",
            "2022-01-12",
            "2022-01-13",
        ],
        dtype="datetime64[ns]",
        freq=None,
    ),
)

X_train_ordinal_diferent_length = pd.DataFrame(
    data=np.array(
        [
            [0.58142898, 0.65138268, 0.12362923, 0.0],
            [0.72969992, 0.58142898, 0.65138268, 0.0],
            [0.97790567, 0.72969992, 0.58142898, 0.0],
            [0.56924731, 0.97790567, 0.72969992, 0.0],
            [0.85369084, 0.56924731, 0.97790567, 0.0],
            [0.75425194, 0.85369084, 0.56924731, 0.0],
            [0.08167939, 0.75425194, 0.85369084, 0.0],
            [0.41262027, 0.20581485, 0.10305721, 1.0],
            [0.82107767, 0.41262027, 0.20581485, 1.0],
            [0.0107816, 0.82107767, 0.41262027, 1.0],
            [0.94951918, 0.0107816, 0.82107767, 1.0],
        ]
    ),
    columns=["lag_1", "lag_2", "lag_3", "_level_skforecast"],
    index=pd.DatetimeIndex(
        [
            "2022-01-07",
            "2022-01-08",
            "2022-01-09",
            "2022-01-10",
            "2022-01-11",
            "2022-01-12",
            "2022-01-13",
            "2022-01-10",
            "2022-01-11",
            "2022-01-12",
            "2022-01-13",
        ],
        dtype="datetime64[ns]",
        freq=None,
    ),
)

X_train_ordinal_category_diferent_length = X_train_ordinal_diferent_length.copy()
X_train_ordinal_category_diferent_length["_level_skforecast"] = (
    X_train_ordinal_category_diferent_length["_level_skforecast"].astype("category")
)


@pytest.mark.parametrize(
    "encoding, X_train",
    [
        ("ordinal", X_train_ordinal),
        ("ordinal_category", X_train_ordinal_category),
        ("onehot", X_train_onehot),
        (None, X_train_ordinal)
    ],
    ids=lambda dt: f"encoding, X_train: {dt}",
)
def test_create_sample_weights_output_using_series_weights(encoding, X_train):
    """
    Test `sample_weights` creation using `series_weights`.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        estimator=LinearRegression(),
        lags=3,
        encoding=encoding,
        series_weights={"series_1": 1.0, "series_2": 2.0},
    )
    forecaster.encoding_mapping_ = {"series_1": 0, "series_2": 1}

    results = forecaster.create_sample_weights(
        series_names_in_=["series_1", "series_2"], X_train=X_train
    )

    expected = np.array(
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0]
    )

    assert np.array_equal(results, expected)


@pytest.mark.parametrize(
    "encoding, X_train",
    [
        ("ordinal", X_train_ordinal),
        ("ordinal_category", X_train_ordinal_category),
        ("onehot", X_train_onehot),
        (None, X_train_ordinal)
    ],
    ids=lambda dt: f"encoding, X_train: {dt}",
)
def test_create_sample_weights_output_using_weight_func(encoding, X_train):
    """
    Test `sample_weights` creation using `weight_func`.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        estimator=LinearRegression(),
        lags=3,
        encoding=encoding,
        weight_func=custom_weights,
    )
    forecaster.encoding_mapping_ = {"series_1": 0, "series_2": 1}
    results = forecaster.create_sample_weights(
        series_names_in_=["series_1", "series_2"],
        X_train=X_train,
    )
    expected = np.array([1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1])

    assert np.array_equal(results, expected)


@pytest.mark.parametrize(
    "weight_func, expected",
    [
        (
            {"series_1": custom_weights},
            np.array(
                [1.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
            ),
        ),
        (
            {"series_2": custom_weights_2},
            np.array(
                [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 3.0, 3.0, 3.0, 3.0, 2.0, 2.0, 2.0]
            ),
        ),
        (
            {"series_1": custom_weights, "series_2": custom_weights_2},
            np.array([1, 0, 0, 0, 1, 1, 1, 3, 3, 3, 3, 2, 2, 2]),
        ),
    ],
    ids=lambda values: f"levels: {values}",
)
def test_create_sample_weights_output_using_weight_func_dict(weight_func, expected):
    """
    Test `sample_weights` creation using `weight_func`.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     encoding           = "ordinal",
                     transformer_series = StandardScaler(),
                     weight_func        = weight_func
                 )
    forecaster.encoding_mapping_ = {"series_1": 0, "series_2": 1}
    results = forecaster.create_sample_weights(
        series_names_in_=["series_1", "series_2"], X_train=X_train_ordinal
    )

    assert np.array_equal(results, expected)


@pytest.mark.parametrize(
    "weight_func, expected",
    [
        (
            {"series_1": custom_weights},
            np.array([1., 0., 0., 0., 1., 1., 1., 1., 1., 1., 1.]),
        ),
        (
            {"series_2": custom_weights_2},
            np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 3.0, 2.0, 2.0, 2.0]),
        ),
        (
            {"series_1": custom_weights, "series_2": custom_weights_2},
            np.array([1, 0, 0, 0, 1, 1, 1, 3, 2, 2, 2]),
        ),
    ],
    ids=lambda values: f"levels: {values}",
)
def test_create_sample_weights_output_using_weight_func_dict_different_series_lengths(
    weight_func, expected
):
    """
    Test `sample_weights` creation using `weight_func` with series of different lengths.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     encoding           = "ordinal",
                     transformer_series = StandardScaler(),
                     weight_func        = weight_func
                 )
    forecaster.encoding_mapping_ = {"series_1": 0, "series_2": 1}
    results = forecaster.create_sample_weights(
        series_names_in_=["series_1", "series_2"],
        X_train=X_train_ordinal_diferent_length,
    )

    assert np.array_equal(results, expected)


@pytest.mark.parametrize(
    "encoding, X_train",
    [
        ("ordinal", X_train_ordinal),
        ("ordinal_category", X_train_ordinal_category),
        ("onehot", X_train_onehot),
        (None, X_train_ordinal)
    ],
    ids=lambda dt: f"encoding, X_train: {dt}",
)
def test_create_sample_weights_output_using_series_weights_and_weight_func(
    encoding, X_train
):
    """
    Test `sample_weights` creation using `series_weights` and `weight_func`.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     encoding           = encoding,
                     transformer_series = StandardScaler(),
                     series_weights     = {"series_1": 1.0, "series_2": 2.0},
                     weight_func        = custom_weights
                 )
    forecaster.encoding_mapping_ = {"series_1": 0, "series_2": 1}
    results = forecaster.create_sample_weights(
        series_names_in_=["series_1", "series_2"], X_train=X_train
    )

    expected = np.array([1, 0, 0, 0, 1, 1, 1, 2, 0, 0, 0, 2, 2, 2], dtype=float)

    assert np.array_equal(results, expected)


@pytest.mark.parametrize(
    "encoding, X_train",
    [
        ("ordinal", X_train_ordinal_diferent_length),
        ("ordinal_category", X_train_ordinal_category_diferent_length),
        ("onehot", X_train_onehot_diferent_length),
        (None, X_train_ordinal_diferent_length)
    ],
    ids=lambda dt: f"encoding, X_train: {dt}",
)
def test_create_sample_weights_output_using_series_weights_and_weight_func_different_series_lengths(
    encoding, X_train
):
    """
    Test `sample_weights` creation using `series_weights` and `weight_func`
    with series of different lengths.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     encoding           = encoding,
                     transformer_series = StandardScaler(),
                     series_weights     = {"series_1": 1.0, "series_2": 2.0},
                     weight_func        = custom_weights
                 )
    forecaster.encoding_mapping_ = {"series_1": 0, "series_2": 1}
    results = forecaster.create_sample_weights(
        series_names_in_=["series_1", "series_2"], X_train=X_train
    )
    
    expected = np.array([1, 0, 0, 0, 1, 1, 1, 0, 2, 2, 2], dtype=float)

    assert np.array_equal(results, expected)


@pytest.mark.parametrize(
    "encoding, X_train",
    [
        ("ordinal", X_train_ordinal),
        ("ordinal_category", X_train_ordinal_category),
        ("onehot", X_train_onehot),
        (None, X_train_ordinal)
    ],
    ids=lambda dt: f"encoding, X_train: {dt}",
)
def test_create_sample_weights_ValueError_when_weights_has_nan(encoding, X_train):
    """
    Test sample_weights ValueError when sample_weight contains NaNs.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     encoding           = encoding,
                     transformer_series = StandardScaler(),
                     weight_func        = custom_weights_nan
                 )
    forecaster.encoding_mapping_ = {"series_1": 0, "series_2": 1}

    err_msg = re.escape("The resulting `weights` cannot have NaN values.")
    with pytest.raises(ValueError, match=err_msg):
        forecaster.create_sample_weights(
            series_names_in_=["series_1", "series_2"], X_train=X_train
        )


@pytest.mark.parametrize(
    "encoding, X_train",
    [
        ("ordinal", X_train_ordinal),
        ("ordinal_category", X_train_ordinal_category),
        ("onehot", X_train_onehot),
        (None, X_train_ordinal)
    ],
    ids=lambda dt: f"encoding, X_train: {dt}",
)
def test_create_sample_weights_ValueError_when_weights_has_negative_values(
    encoding, X_train
):
    """
    Test sample_weights ValueError when sample_weight contains negative values.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     encoding           = encoding,
                     transformer_series = StandardScaler(),
                     weight_func        = custom_weights_negative
                 )
    forecaster.encoding_mapping_ = {"series_1": 0, "series_2": 1}

    err_msg = re.escape("The resulting `weights` cannot have negative values.")
    with pytest.raises(ValueError, match=err_msg):
        forecaster.create_sample_weights(
            series_names_in_=["series_1", "series_2"], X_train=X_train
        )


@pytest.mark.parametrize(
    "encoding, X_train",
    [
        ("ordinal", X_train_ordinal),
        ("ordinal_category", X_train_ordinal_category),
        ("onehot", X_train_onehot),
        (None, X_train_ordinal)
    ],
    ids=lambda dt: f"encoding, X_train: {dt}",
)
def test_create_sample_weights_ValueError_when_weights_all_zeros(encoding, X_train):
    """
    Test sample_weights ValueError when the sum of the weights is zero.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     encoding           = encoding,
                     transformer_series = StandardScaler(),
                     series_weights     = {"series_1": 0, "series_2": 0},
                 )
    forecaster.encoding_mapping_ = {"series_1": 0, "series_2": 1}

    err_msg = re.escape(
        "The resulting `weights` cannot be normalized because "
        "the sum of the weights is zero."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster.create_sample_weights(
            series_names_in_=["series_1", "series_2"], X_train=X_train
        )



# Unit test predict
# ==============================================================================
from pathlib import Path
import pytest
import joblib
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from ..._population_drift import PopulationDriftDetector

# Fixtures
THIS_DIR = Path(__file__).parent
data = joblib.load(THIS_DIR/'fixture_data_population_drift.joblib')
results_nannyml = joblib.load(THIS_DIR/'fixture_results_nannyml.joblib')
results_multiseries = joblib.load(THIS_DIR/'fixture_results_multiseries.joblib')
summary_multiseries = joblib.load(THIS_DIR/'fixture_summary_multiseries.joblib')


# NOTE: Code used to generate fixture_results_nannyml
# detector = nml.UnivariateDriftCalculator(
#     column_names=data_train.columns.tolist(),
#     timestamp_column_name="date_time",
#     chunk_period='M',
#     categorical_methods=['chi2', 'jensen_shannon'],
#     continuous_methods=['kolmogorov_smirnov', 'jensen_shannon'],
# )
# detector.fit(reference_data=data_train.reset_index())
# results_nannyml = detector.calculate(data=data_new.reset_index())
# results_nannyml = results_nannyml.filter(period='analysis').to_df(multilevel=False)


def test_predict_NotFittedError_when_detector_not_fitted():
    """
    Test NotFittedError is raised when trying to predict before fitting the detector.
    """
    detector = PopulationDriftDetector(
        chunk_size='ME',            
        threshold=0.99
    )
    X = data
    err_msg = (
        "This PopulationDriftDetector instance is not fitted yet. "
        "Call 'fit' with appropriate arguments before using this estimator."
    )
    with pytest.raises(NotFittedError, match=err_msg):
        detector.predict(X=X)


def test_predict_ValueError_when_X_not_dataframe():
    """
    Test ValueError is raised when X is not a pandas DataFrame.
    """
    detector = PopulationDriftDetector(
        chunk_size='ME',            
        threshold=0.99
    )
    detector.fit(data)
    X = 'not a dataframe'
    err_msg = f"`X` must be a pandas DataFrame. Got {type(X)} instead."
    with pytest.raises(ValueError, match=err_msg):
        detector.predict(X=X)


def test_predict_ValueError_when_chunk_size_is_DateOffset_but_X_index_not_datetimeindex():
    """
    Test ValueError is raised when chunk_size is a pandas DateOffset but X does 
    not have a DatetimeIndex.
    """
    detector = PopulationDriftDetector(
        chunk_size='ME',            
        threshold=0.99
    )
    detector.fit(data)
    X = data.reset_index()
    err_msg = "`chunk_size` is a pandas DateOffset but `X` does not have a DatetimeIndex."
    with pytest.raises(ValueError, match=err_msg):
        detector.predict(X=X)


def test_predict_output_equivalence_nannyml():
    """
    Test that the output of PopulationDriftDetector.predict is equivalent to
    the output of NannyML's univariate drift detection for each series_id and feature.
    """
    data_train = data.iloc[: len(data) // 2].copy()
    data_new  = data.iloc[len(data) // 2 :].copy()
    data_train['weather'] = data_train['weather'].astype('category')
    data_new['weather'] = pd.Categorical(data_new['weather'], categories=data_train['weather'].cat.categories)

    detector = PopulationDriftDetector(
        chunk_size='ME',            
        threshold=0.99
    )
    detector.fit(data_train)
    results_skforecast, _ = detector.predict(data_new)

    features = results_skforecast["feature"].unique()
    for feature in features:
        print(f"Feature: {feature}")
        cols = [
            "chunk_index",
            "chunk_start_date",
            "chunk_end_date",
            f"{feature}_kolmogorov_smirnov_value",
            f"{feature}_jensen_shannon_value",
            f"{feature}_chi2_value",
        ]
        # select columns if they exist in results_skforecast
        cols = [col for col in cols if col in results_nannyml.columns]
        df_nannyml = results_nannyml[cols].copy()
        df_nannyml = df_nannyml.rename(
            columns={
                "chunk_index": "chunk",
                "chunk_start_date": "chunk_start",
                "chunk_end_date": "chunk_end",
                f"{feature}_kolmogorov_smirnov_value": "ks_statistic",
                f"{feature}_jensen_shannon_value": "jensen_shannon",
                f"{feature}_chi2_value": "chi2_statistic",
            }
        )
        df_nannyml["feature"] = feature
        if "chi2_statistic" not in df_nannyml.columns:
            df_nannyml["chi2_statistic"] = np.nan
        if "ks_statistic" not in df_nannyml.columns:
            df_nannyml["ks_statistic"] = np.nan
        if "jensen_shannon" not in df_nannyml.columns:
            df_nannyml["jensen_shannon"] = np.nan

        df_nannyml = df_nannyml[
            [
                "chunk",
                "chunk_start",
                "chunk_end",
                "feature",
                "ks_statistic",
                "chi2_statistic",
                "jensen_shannon",
            ]
        ]

        df_skforecast = results_skforecast.query(f"feature == '{feature}'")[
            [
                "chunk",
                "chunk_start",
                "chunk_end",
                "feature",
                "ks_statistic",
                "chi2_statistic",
                "jensen_shannon",
            ]
        ]

        df_all = pd.merge(
            df_nannyml,
            df_skforecast,
            on=["chunk", "chunk_start", "feature"],
            suffixes=("_nannyml", "_skforecast"),
        )
        pd.testing.assert_series_equal(
            df_all["ks_statistic_nannyml"],
            df_all["ks_statistic_skforecast"],
            check_names=False,
        )
        pd.testing.assert_series_equal(
            df_all["chi2_statistic_nannyml"],
            df_all["chi2_statistic_skforecast"],
            check_names=False,
        )
        pd.testing.assert_series_equal(
            df_all["jensen_shannon_nannyml"],
            df_all["jensen_shannon_skforecast"],
            check_names=False,
        )


def test_predict_output_when_multiple_series():
    """
    Test that PopulationDriftDetector.predict works when data contains multiple series.
    """

    data_multiseries = pd.concat(
        [
            data.assign(series='series_1'),
            data.assign(series='series_2'),
            data.assign(series='series_3')
        ]
    ).set_index('series', append=True).swaplevel(0, 1)

    detector = PopulationDriftDetector(
        chunk_size='ME',            
        threshold=0.95
    )
    detector.fit(data_multiseries)
    results, summary = detector.predict(data_multiseries)
    pd.testing.assert_frame_equal(results, results_multiseries)
    pd.testing.assert_frame_equal(summary, summary_multiseries)
    


# Unit test preprocess_levels_self_last_window_multiseries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.exceptions import IgnoredArgumentWarning
from skforecast.utils import preprocess_levels_self_last_window_multiseries


def test_preprocess_levels_self_last_window_multiseries_IgnoredArgumentWarning_when_not_available_self_last_window_for_some_levels():
    """
    Test IgnoredArgumentWarning is raised when last window is not available for 
    levels because it was not stored during fit.
    """

    levels = ['1', '2', '3']
    input_levels_is_list = False
    last_window_ = {
        '1': pd.Series(np.arange(10), name='1'),
        '3': pd.Series(np.arange(10), name='3'),
    }

    warn_msg = re.escape(
        ("Levels {'2'} are excluded from prediction "
         "since they were not stored in `last_window_` attribute "
         "during training. If you don't want to retrain the "
         "Forecaster, provide `last_window` as argument.")
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        levels, last_window = preprocess_levels_self_last_window_multiseries(
                                  levels               = levels,
                                  input_levels_is_list = input_levels_is_list, 
                                  last_window_         = last_window_
                              )

    expected = (
        ['1', '3'],
        pd.DataFrame({'1': np.arange(10), '3': np.arange(10)})
    )
    
    assert levels == expected[0]
    pd.testing.assert_frame_equal(last_window, expected[1])


@pytest.mark.parametrize("last_window_",
                         [{'1': pd.Series(np.arange(10), name='1')}, None],
                         ids=lambda lw: f"last_window: {type(lw)}")
def test_preprocess_levels_self_last_window_multiseries_ValueError_when_not_last_window_for_levels(last_window_):
    """
    Test ValueError is raised when last window is not available for all 
    levels because it was not stored during fit.
    """
    
    levels = ['2']
    input_levels_is_list = False

    err_msg = re.escape(
        "No series to predict. None of the series {'2'} are present in "
        "`last_window_` attribute. Provide `last_window` as argument "
        "in predict method."
    )
    with pytest.raises(ValueError, match = err_msg):
        levels, last_window = preprocess_levels_self_last_window_multiseries(
                                  levels               = levels,
                                  input_levels_is_list = input_levels_is_list, 
                                  last_window_         = last_window_
                              )


def test_preprocess_levels_self_last_window_multiseries_IgnoredArgumentWarning_when_levels_is_list_and_different_last_index_in_self_last_window_DatetimeIndex():
    """
    Test IgnoredArgumentWarning is raised when levels is a list and have 
    different last index in last_window_ attribute using a DatetimeIndex.
    """

    levels = ['1', '2']
    input_levels_is_list = True
    last_window_ = {
        '1': pd.Series(np.arange(10), name='1'),
        '2': pd.Series(np.arange(5), name='2'),
    }
    last_window_['1'].index = pd.date_range(start='2020-01-01', periods=10)
    last_window_['2'].index = pd.date_range(start='2020-01-01', periods=5)

    expected = (
        ['1'],
        pd.DataFrame({'1': np.arange(10)},
                     index=pd.date_range(start='2020-01-01', periods=10))
    )
    
    warn_msg = re.escape(
        ("Only series whose last window ends at the same index "
         "can be predicted together. Series that do not reach the "
         "maximum index, '2020-01-10 00:00:00', are excluded "
         "from prediction: {'2'}.")
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        levels, last_window = preprocess_levels_self_last_window_multiseries(
                                  levels               = levels,
                                  input_levels_is_list = input_levels_is_list, 
                                  last_window_         = last_window_
                              )

    assert levels == expected[0]
    pd.testing.assert_frame_equal(last_window, expected[1])


@pytest.mark.parametrize("input_levels_is_list", 
                         [True, False], 
                         ids=lambda levels: f'input_levels_is_list: {levels}')
def test_output_preprocess_levels_self_last_window_multiseries(input_levels_is_list):
    """
    Test output preprocess_levels_self_last_window_multiseries for different inputs.
    """

    levels = ['1', '2', '3']
    last_window_ = {
        '1': pd.Series(np.arange(10), name='1'),
        '2': pd.Series(np.arange(10, 20), name='2'),
        '3': pd.Series(np.arange(20, 30), name='3'),
    }
    last_window_['1'].index = pd.date_range(start='2020-01-01', periods=10)
    last_window_['2'].index = pd.date_range(start='2020-01-01', periods=10)
    last_window_['3'].index = pd.date_range(start='2020-01-01', periods=10)

    expected = (
        ['1', '2', '3'],
        pd.DataFrame(
            {'1': np.arange(10),
             '2': np.arange(10, 20),
             '3': np.arange(20, 30)},
             index=pd.date_range(start='2020-01-01', periods=10))
    )

    levels, last_window = preprocess_levels_self_last_window_multiseries(
                              levels               = levels,
                              input_levels_is_list = input_levels_is_list, 
                              last_window_         = last_window_
                          )

    assert levels == expected[0]
    pd.testing.assert_frame_equal(last_window, expected[1])



# Unit test fit ForecasterRecursive
# ==============================================================================
import pytest
from pytest import approx
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from skforecast.preprocessing import RollingFeatures
from skforecast.recursive import ForecasterRecursive

# Fixtures
from .fixtures_forecaster_recursive import y
from .fixtures_forecaster_recursive import exog


def custom_weights(index):  # pragma: no cover
    """
    Return 0 if index is between 20 and 40 else 1.
    """
    weights = np.where(
                (index >= 20) & (index <= 40),
                0,
                1
              )
    
    return weights


def test_forecaster_y_exog_features_stored():
    """
    Test forecaster stores y and exog features after fitting.
    """
    rolling = RollingFeatures(
        stats=['ratio_min_max', 'median'], window_sizes=4
    )
    forecaster = ForecasterRecursive(
        LinearRegression(), lags=3, window_features=rolling
    )
    forecaster.fit(y=y, exog=exog)

    series_name_in_ = 'y'
    exog_in_ = True
    exog_type_in_ = type(exog)
    exog_names_in_ = ['exog']
    exog_dtypes_in_ = {'exog': exog.dtype}
    exog_dtypes_out_ = {'exog': exog.dtype}
    X_train_window_features_names_out_ = ['roll_ratio_min_max_4', 'roll_median_4']
    X_train_exog_names_out_ = ['exog']
    X_train_features_names_out_ = [
        'lag_1', 'lag_2', 'lag_3', 'roll_ratio_min_max_4', 'roll_median_4', 'exog'
    ]
    
    assert forecaster.series_name_in_ == series_name_in_
    assert forecaster.exog_in_ == exog_in_
    assert forecaster.exog_type_in_ == exog_type_in_
    assert forecaster.exog_names_in_ == exog_names_in_
    assert forecaster.exog_dtypes_in_ == exog_dtypes_in_
    assert forecaster.exog_dtypes_out_ == exog_dtypes_out_
    assert forecaster.X_train_window_features_names_out_ == X_train_window_features_names_out_
    assert forecaster.X_train_exog_names_out_ == X_train_exog_names_out_
    assert forecaster.X_train_features_names_out_ == X_train_features_names_out_


def test_forecaster_DatetimeIndex_index_freq_stored():
    """
    Test serie_with_DatetimeIndex.index.freq is stored in forecaster.index_freq.
    """
    serie_with_DatetimeIndex = pd.Series(
        data  = [1, 2, 3, 4, 5],
        index = pd.date_range(start='2022-01-01', periods=5)
    )
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=serie_with_DatetimeIndex)
    expected = serie_with_DatetimeIndex.index.freq
    results = forecaster.index_freq_

    assert results == expected


def test_forecaster_index_step_stored():
    """
    Test serie without DatetimeIndex, step is stored in forecaster.index_freq.
    """
    y = pd.Series(data=np.arange(10))
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=y)
    expected = y.index.step
    results = forecaster.index_freq_

    assert results == expected


def test_fit_in_sample_residuals_stored():
    """
    Test that values of in_sample_residuals_ are stored after fitting.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(5)), store_in_sample_residuals=True)
    results = forecaster.in_sample_residuals_
    expected = np.array([0., 0.])

    assert isinstance(results, np.ndarray)
    np.testing.assert_array_almost_equal(results, expected)


def test_fit_same_residuals_when_residuals_greater_than_10000():
    """
    Test fit return same residuals when residuals len is greater than 10_000.
    Testing with two different forecaster.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(12_000)), store_in_sample_residuals=True)
    results_1 = forecaster.in_sample_residuals_
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(12_000)), store_in_sample_residuals=True)
    results_2 = forecaster.in_sample_residuals_
    
    assert isinstance(results_1, np.ndarray)
    assert isinstance(results_2, np.ndarray)
    assert len(results_1) == 10_000
    assert len(results_2) == 10_000
    np.testing.assert_array_almost_equal(results_1, results_2)


def test_fit_in_sample_residuals_by_bin_stored():
    """
    Test that values of in_sample_residuals_by_bin are stored after fitting.
    """
    forecaster = ForecasterRecursive(
                     estimator     = LinearRegression(),
                     lags          = 5,
                     binner_kwargs = {'n_bins': 3}
                 )
    forecaster.fit(y, store_in_sample_residuals=True)

    X_train, y_train = forecaster.create_train_X_y(y)
    forecaster.estimator.fit(X_train, y_train)
    predictions_estimator = forecaster.estimator.predict(X_train)
    expected_1 = y_train - predictions_estimator

    expected_2 = {
        0: np.array([
                0.0334789 , -0.12428472,  0.34053202, -0.40668544, -0.29246428,
                0.16990408, -0.02118736, -0.24234062, -0.11745596,  0.1697826 ,
                -0.01432662, -0.00063421, -0.03462192,  0.41322689,  0.19077889
            ]),
        1: np.array([
                -0.07235524, -0.10880301, -0.07773704, -0.09070227,  0.21559424,
                -0.29380582,  0.03359274,  0.10109702,  0.2080735 , -0.17086244,
                0.01929597, -0.09396861, -0.0670198 ,  0.38248168, -0.01100463
            ]),
        2: np.array([
                0.44780048,  0.03560524, -0.04960603,  0.24323339,  0.12651656,
                -0.46533293, -0.17532266, -0.24111645,  0.3805961 , -0.05842153,
                0.08927473, -0.42295249, -0.32047616,  0.38902396, -0.01640072
            ])
    }

    expected_3 = {
        0: (0.31791969404305154, 0.47312737276420375),
        1: (0.47312737276420375, 0.5259220171775293),
        2: (0.5259220171775293, 0.6492244994657664)
    }

    np.testing.assert_array_almost_equal(
        np.sort(forecaster.in_sample_residuals_),
        np.sort(expected_1)
    )
    for k in expected_2.keys():
        np.testing.assert_array_almost_equal(forecaster.in_sample_residuals_by_bin_[k], expected_2[k])
    for k in expected_3.keys():
        assert forecaster.binner_intervals_[k][0] == approx(expected_3[k][0])
        assert forecaster.binner_intervals_[k][1] == approx(expected_3[k][1])


def test_fit_in_sample_residuals_not_stored_probabilistic_mode_binned():
    """
    Test that values of in_sample_residuals_ are not stored after fitting
    when `store_in_sample_residuals=False`. Binner intervals are stored.
    """
    forecaster = ForecasterRecursive(
                     estimator     = LinearRegression(),
                     lags          = 5,
                     binner_kwargs = {'n_bins': 3}
                 )
    forecaster.fit(y, store_in_sample_residuals=False)

    expected_binner_intervals_ = {
        0: (0.31791969404305154, 0.47312737276420375),
        1: (0.47312737276420375, 0.5259220171775293),
        2: (0.5259220171775293, 0.6492244994657664)
    }

    assert forecaster.in_sample_residuals_ is None
    assert forecaster.in_sample_residuals_by_bin_ is None
    assert forecaster.binner_intervals_.keys() == expected_binner_intervals_.keys()
    for k in expected_binner_intervals_.keys():
        assert forecaster.binner_intervals_[k][0] == approx(expected_binner_intervals_[k][0])
        assert forecaster.binner_intervals_[k][1] == approx(expected_binner_intervals_[k][1])


def test_fit_in_sample_residuals_not_stored_probabilistic_mode_False():
    """
    Test that values of in_sample_residuals_ are not stored after fitting
    when `store_in_sample_residuals=False` and _probabilistic_mode=False.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster._probabilistic_mode = False
    forecaster.fit(y=pd.Series(np.arange(10), name='y'), store_in_sample_residuals=False)

    assert forecaster.in_sample_residuals_ is None
    assert forecaster.in_sample_residuals_by_bin_ is None
    assert forecaster.binner_intervals_ is None


@pytest.mark.parametrize("store_last_window", 
                         [True, False], 
                         ids=lambda lw: f'store_last_window: {lw}')
def test_fit_last_window_stored(store_last_window):
    """
    Test that values of last window are stored after fitting.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(50)), store_last_window=store_last_window)
    expected = pd.DataFrame(np.array([47, 48, 49]), index=[47, 48, 49], columns=['y'])

    if store_last_window:
        pd.testing.assert_frame_equal(forecaster.last_window_, expected)
    else:
        assert forecaster.last_window_ is None


def test_fit_model_coef_when_using_weight_func():
    """
    Check the value of the estimator coefs when using a `weight_func`.
    """
    forecaster = ForecasterRecursive(
                     estimator   = LinearRegression(),
                     lags        = 5,
                     weight_func = custom_weights
                 )
    forecaster.fit(y=y)
    results = forecaster.estimator.coef_
    expected = np.array([0.01211677, -0.20981367,  0.04214442, -0.0369663, -0.18796105])

    np.testing.assert_almost_equal(results, expected)


def test_fit_model_coef_when_not_using_weight_func():
    """
    Check the value of the estimator coefs when not using a `weight_func`.
    """
    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=5)
    forecaster.fit(y=y)
    results = forecaster.estimator.coef_
    expected = np.array([0.16773502, -0.09712939,  0.10046413, -0.09971515, -0.15849756])

    np.testing.assert_almost_equal(results, expected)
